from jolt import attributes, Parameter, Task
from jolt.pkgs import llvm
from jolt.tasks import TaskRegistry


@attributes.requires("requires_llvm")
class Clang(Task):
    name = "clang"
    version = Parameter(llvm.VERSION, help="Clang version.")

    requires_llvm = ["llvm:version={version}"]

    def publish(self, artifact, tools):
        artifact.environ.LLVM_TOOL_clang = "clang"
        artifact.environ.LLVM_TOOL_clangpp = "clang++"
        artifact.environ.AR = "llvm-ar"
        artifact.environ.CC = "clang"
        artifact.environ.CXX = "clang++"
        artifact.environ.LD = "clang++"
        artifact.environ.NM = "llvm-nm"
        artifact.environ.OBJCOPY = "llvm-objcopy"
        artifact.environ.OBJDUMP = "llvm-objdump"
        artifact.environ.PATH.append("bin")
        artifact.environ.RANLIB = "llvm-ranlib"
        artifact.environ.STRIP = "llvm-strip"


class ClangFormat(llvm.LLVMFromSrc):
    name = "clang-format"
    projects = ["clang", "clang-tools-extra"]
    targets = []
    install = ["install-clang-format"]


class ClangTidy(llvm.LLVMFromSrc):
    name = "clang-tidy"
    projects = ["clang", "clang-tools-extra"]
    targets = []
    install = ["install-clang-tidy"]


TaskRegistry.get().add_task_class(Clang)
TaskRegistry.get().add_task_class(ClangFormat)
TaskRegistry.get().add_task_class(ClangTidy)



from jolt import attributes, Alias, BooleanParameter, Parameter, Task
from jolt.pkgs import libffi, sqlite, ssl, zlib, zstd, xz
from jolt.plugins import autotools, git, pkgconfig
from jolt.tasks import TaskRegistry


def _unpack_adjust_scripts(artifact, tools):
    bindir = "Scripts" if artifact.strings.system == "windows" else "bin"

    with tools.cwd(artifact.path, bindir):
        # Adjust paths in scripts
        for script in tools.glob("*"):
            # Ignore python executables
            if script.startswith("python"):
                continue
            tools.replace_in_file(script, artifact.strings.install_prefix, artifact.final_path)

    artifact.strings.install_prefix = artifact.final_path


@attributes.requires("requires_git")
@attributes.requires("requires_libffi_{libffi[on,off]}")
@attributes.requires("requires_sqlite_{sqlite[on,off]}")
@attributes.requires("requires_ssl_{ssl[on,off]}")
@attributes.requires("requires_zlib_{zlib[on,off]}")
@attributes.requires("requires_zstd_{zstd[on,off]}")
@attributes.requires("requires_xz_{xz[on,off]}")
@pkgconfig.requires()
class CPythonPosix(autotools.Autotools):
    """ Builds and publishes CPython libraries and headers. """

    name = "cpython/posix"
    version = Parameter("3.14.2", help="CPython version.")
    libffi = BooleanParameter(True, help="Enable libffi support")
    sqlite = BooleanParameter(True, help="Enable SQLite support")
    ssl = BooleanParameter(True, help="Enable SSL support")
    zlib = BooleanParameter(True, help="Enable zlib support")
    zstd = BooleanParameter(True, help="Enable zstd support")
    xz = BooleanParameter(True, help="Enable xz support")
    requires_git = ["git:url=https://github.com/python/cpython.git,rev=v{version}"]
    requires_libffi_on = ["libffi"]
    requires_sqlite_on = ["sqlite"]
    requires_ssl_on = ["virtual/ssl"]
    requires_zlib_on = ["virtual/zlib"]
    requires_zstd_on = ["zstd"]
    requires_xz_on = ["xz:pic=true"]
    srcdir = "{git[cpython]}"
    options = [
        "--enable-optimizations",
        "--with-ensurepip=install",
    ]

    @property
    def version_major(self):
        ver = str(self.version).split(".")
        return f"{ver[0]}.{ver[1]}"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.strings.version = str(self.version)
        artifact.strings.version_major = str(self.version_major)

    def unpack(self, artifact, tools):
        super().unpack(artifact, tools)
        _unpack_adjust_scripts(artifact, tools)


@attributes.requires("requires_git")
@attributes.requires("requires_ssl_{ssl[on,off]}")
@attributes.common_metadata()
class CPythonWin32(Task):
    """ Builds and publishes CPython libraries and headers. """

    name = "cpython/win32"
    version = Parameter("3.14.2", help="CPython version.")
    ssl = BooleanParameter(True, help="Enable SSL support")
    requires_git = ["git:url=https://github.com/python/cpython.git,rev=v{version}"]
    requires_ssl_on = ["virtual/ssl"]
    srcdir = "{git[cpython]}"

    @property
    def version_major(self):
        ver = str(self.version).split(".")
        return f"{ver[0]}.{ver[1]}"

    @property
    def version_major_compact(self):
        ver = str(self.version).split(".")
        return f"{ver[0]}{ver[1]}"

    def run(self, deps, tools):
        with tools.cwd(self.srcdir, "PCbuild"):
            tools.run("build.bat")

    def publish(self, artifact, tools):
        artifact.environ.CMAKE_PREFIX_PATH.append(".")
        artifact.strings.version = str(self.version)
        artifact.strings.version_major = str(self.version_major)
        with tools.cwd(self.srcdir):
            artifact.collect("Include", "include")
            artifact.collect("Lib", "lib")
        with tools.cwd(self.srcdir, "PCbuild", "amd64"):
            artifact.collect("python.exe", "bin/python3.exe")
            artifact.collect("python.exe", "bin/python{version_major}.exe")
            artifact.collect("*.dll", "bin/")
            artifact.collect("*.lib", "lib/")
        with tools.tmpdir() as tmp, tools.cwd(tmp):
            tools.write_file(
                "python3.pc",
                """
# See: man pkg-config
prefix=${{pcfiledir}}/../..
exec_prefix=${{prefix}}/bin
libdir=${{prefix}}/lib
includedir=${{prefix}}/include

Name: Python
Description: Build a C extension for Python
Version: {version}-{identity}
Libs: -L${{libdir}} -lpython3
Cflags: -I${{includedir}}
""")
            tools.write_file(
                "python3-embed.pc",
                """
# See: man pkg-config
prefix=${{pcfiledir}}/../..
exec_prefix=${{prefix}}/bin
libdir=${{prefix}}/lib
includedir=${{prefix}}/include

Name: Python
Description: Embed Python into an application
Version: {version}-{identity}
Libs: -L${{libdir}} -lpython{version_major_compact}
Cflags: -I${{includedir}}
""")
            artifact.collect("*.pc", "lib/pkgconfig/")

    def unpack(self, artifact, tools):
        super().unpack(artifact, tools)
        _unpack_adjust_scripts(artifact, tools)


@attributes.requires("requires_{system}")
@attributes.system
class CPython(Alias):
    """ Alias for CPython """
    name = "cpython"
    version = Parameter("3.14.2", help="CPython version.")
    requires_darwin = ["cpython/posix:version={version}"]
    requires_linux = requires_darwin
    requires_windows = ["cpython/win32:version={version}"]


@pkgconfig.to_cxxinfo("python3-embed")
@attributes.common_metadata()
@attributes.system
@pkgconfig.requires()
class CPythonEmbed(Task):
    name = "cpython/embed"
    requires = ["cpython"]
    selfsustained = True

    def run(self, deps, tools):
        name = "cpython/win32" if self.system == "windows" else "cpython/posix"
        self.cpython = deps[name]

    def publish(self, artifact, tools):
        with tools.cwd(self.cpython.path):
            artifact.collect("*", symlinks=True)


@pkgconfig.to_cxxinfo("python3")
@attributes.common_metadata()
@attributes.system
@pkgconfig.requires()
class CPythonExtend(Task):
    name = "cpython/extend"
    requires = ["cpython"]
    selfsustained = True

    def run(self, deps, tools):
        name = "cpython/win32" if self.system == "windows" else "cpython/posix"
        self.cpython = deps[name]

    def publish(self, artifact, tools):
        with tools.cwd(self.cpython.path):
            artifact.collect("*", symlinks=True)


TaskRegistry.get().add_task_class(CPythonPosix)
TaskRegistry.get().add_task_class(CPythonWin32)
TaskRegistry.get().add_task_class(CPython)
TaskRegistry.get().add_task_class(CPythonEmbed)
TaskRegistry.get().add_task_class(CPythonExtend)



import os
from jolt import attributes, Parameter
from jolt.plugins import git
from jolt.tasks import Task, TaskRegistry


@attributes.common_metadata()
@attributes.requires("requires_git")
@attributes.system
class B2(Task):
    name = "b2"
    version = Parameter("5.4.2", help="B2 version.")
    requires_git = ["git:url=https://github.com/bfgroup/b2.git"]

    def run(self, artifact, tools):
        self.builddir = tools.builddir("build", incremental=True)
        self.installdir = tools.builddir("install")
        with tools.cwd("{git[b2]}"):
            if self.system == "windows":
                tools.run(".\\bootstrap.bat")
                tools.run(".\\b2 install --prefix={installdir} --build-dir={builddir} -j{}", tools.cpu_count())
            else:
                tools.run("./bootstrap.sh")
                tools.run("./b2 install --prefix={installdir} --build-dir={builddir} -j{}", tools.cpu_count())

    def publish(self, artifact, tools):
        with tools.cwd(self.installdir):
            artifact.collect("*", symlinks=True)


TaskRegistry.get().add_task_class(B2)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["cares"])
class CAres(cmake.CMake):
    name = "c-ares"
    version = Parameter("1.34.6", help="c-ares version.")
    pic = BooleanParameter(False, help="Build with position independent code.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/c-ares/c-ares.git,rev=v{version}"]
    srcdir = "{git[c-ares]}"
    options = [
        "CARES_SHARED={shared[ON,OFF]}",
        "CARES_STATIC={shared[OFF,ON]}",
        "CARES_BUILD_TESTS=OFF",
        "CMAKE_POSITION_INDEPENDENT_CODE={pic[ON,OFF]}",
    ]


TaskRegistry.get().add_task_class(CAres)



from jolt import attributes, Parameter, Task
from jolt.plugins import git, autotools
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_texinfo")
@autotools.requires(automake=False, libtool=False)
class Automake(autotools.Autotools):
    name = "automake"
    version = Parameter("1.18.1", help="Automake version.")
    requires_git = ["fetch:alias=src,url=https://ftpmirror.gnu.org/gnu/automake/automake-{version}.tar.xz"]
    requires_texinfo = ["texinfo"]
    srcdir = "{fetch[src]}/automake-{version}"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        self.unpack(artifact, tools)

    def unpack(self, artifact, tools):
        with tools.cwd(artifact.path, "bin"):
            for file in tools.glob("*"):
                tools.replace_in_file(
                    file,
                    artifact.strings.install_prefix,
                    artifact.final_path)

        for file in tools.glob(artifact.path + "/share/automake-*/Automake/Config.pm"):
            tools.replace_in_file(
                file,
                artifact.strings.install_prefix,
                artifact.final_path)

        with tools.cwd(artifact.path):
            for file in tools.glob("share/aclocal-*"):
                artifact.environ.ACLOCAL_PATH.append(file)

        artifact.strings.install_prefix = artifact.final_path


TaskRegistry.get().add_task_class(Automake)



from jolt import attributes, BooleanParameter, Parameter
from jolt.plugins import cmake, cxxinfo, git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_src")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["expat"])
class Libexpat(cmake.CMake):
    name = "libexpat"
    version = Parameter("2.7.3", help="Expat version.")
    shared = BooleanParameter(False, help="Build shared libraries")
    version_tag = property(lambda self: str(self.version).replace('.', '_'))
    requires_src = ["fetch:alias=src,url=https://github.com/libexpat/libexpat/releases/download/R_{version_tag}/expat-{version}.tar.gz"]
    srcdir = "{fetch[src]}/expat-{version}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
    ]


TaskRegistry.get().add_task_class(Libexpat)



from jolt import attributes, Parameter
from jolt.plugins import cmake, git, autotools, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@autotools.requires()
@pkgconfig.requires()
@pkgconfig.to_cxxinfo("asio")
class Asio(autotools.Autotools):
    name = "asio"
    version = Parameter("1.36.0", help="asio version.")
    requires_git = ["git:url=https://github.com/chriskohlhoff/asio.git,rev=asio-{_version}"]
    srcdir = "{git[asio]}/asio"

    @property
    def _version(self):
        return str(self.version).replace(".", "-")

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.CMAKE_PREFIX_PATH.append(".")


TaskRegistry.get().add_task_class(Asio)



from jolt import Alias
from jolt.tasks import TaskRegistry


class GlProto(Alias):
    name = "glproto"
    requires = ["xorg/glproto"]


class Libx11(Alias):
    name = "libx11"
    requires = ["xorg/libx11"]


class Libxext(Alias):
    name = "libxext"
    requires = ["xorg/libxext"]


class Libxshmfence(Alias):
    name = "libxshmfence"
    requires = ["xorg/libxshmfence"]


class LibXcb(Alias):
    name = "libxcb"
    requires = ["xorg/libxcb"]


class LibX11Xcb(Alias):
    name = "libx11-xcb"
    requires = ["xorg/libx11-xcb"]


class Xproto(Alias):
    name = "xproto"
    requires = ["xorg/xproto"]


class LibXrandr(Alias):
    name = "libxrandr"
    requires = ["xorg/libxrandr"]


class LibXxf86vm(Alias):
    name = "libxxf86vm"
    requires = ["xorg/libxxf86vm"]


TaskRegistry.get().add_task_class(GlProto)
TaskRegistry.get().add_task_class(Libx11)
TaskRegistry.get().add_task_class(LibX11Xcb)
TaskRegistry.get().add_task_class(LibXcb)
TaskRegistry.get().add_task_class(Libxext)
TaskRegistry.get().add_task_class(LibXrandr)
TaskRegistry.get().add_task_class(LibXxf86vm)
TaskRegistry.get().add_task_class(Libxshmfence)
TaskRegistry.get().add_task_class(Xproto)



from jolt import attributes, Parameter, Task
from jolt.plugins import cxxinfo, fetch
from jolt.tasks import BooleanParameter, TaskRegistry


@attributes.common_metadata()
@attributes.requires("requires_src")
@cxxinfo.publish(libraries=["cap", "psx"])
class Libcap(Task):
    name = "libcap"
    version = Parameter("2.77", help="Libcap version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_src = ["git:url=https://git.kernel.org/pub/scm/libs/libcap/libcap.git,rev=libcap-{version}"]
    srcdir = "{git[libcap]}"

    def run(self, deps, tools):
        self.installdir = tools.builddir("install")
        with tools.cwd(self.srcdir):
            tools.run("make -j {}", tools.cpu_count())
            tools.run("make install DESTDIR={}", self.installdir)

        # Patch pkgconfig files
        with tools.cwd(self.installdir):
            for pc in tools.glob("lib*/pkgconfig/*.pc"):
                tools.replace_in_file(
                    pc,
                    "prefix=/usr",
                    "prefix=${{pcfiledir}}/../..",
                )
                tools.replace_in_file(
                    pc,
                    "libdir=/",
                    "libdir=${{prefix}}/"
                )
                tools.replace_in_file(
                    pc,
                    "includedir=/",
                    "includedir=${{prefix}}/"
                )

    def publish(self, artifact, tools):
        with tools.cwd(self.installdir):
            if self.shared:
                for lib in tools.glob("lib*/*.a"):
                    tools.unlink(lib)
            else:
                for lib in tools.glob("lib*/*.so*"):
                    tools.unlink(lib)
            artifact.collect("*", symlinks=True)


TaskRegistry.get().add_task_class(Libcap)



import hashlib
from jolt import BooleanParameter, Parameter
from jolt.plugins.git import ErrorDict
from jolt.tasks import Resource, TaskRegistry
from jolt import filesystem as fs
from jolt import utils
from jolt.error import raise_error_if
from jolt.loader import JoltLoader
from jolt.tools import SUPPORTED_ARCHIVE_TYPES


class Fetch(Resource):
    """
    Fetch a file from a URL and extract it.

    By default, the fetched file is extracted if it is a supported archive type.
    Extraction may be disable by assigning the `extract` parameter to false.

    The fetched content is placed in a build directory named after the resource.
    The `path` parameter can be used to specify a different location relative to
    the workspace root.

    The path of the fetched content is made available to consuming tasks through
    the `fetch` attribute. The `fetch` attribute is a dictionary where the key
    is the name of the fetched file and the value is a path relative to the
    workspace root. You can specify an alias for the filename / key through the
    `alias` parameter.

    The plugin must be loaded before it can be used. This is done by importing
    the module, or by adding the following line to the configuration file:

    .. code-block:: ini

            [fetch]


    Example:

    .. code-block:: python

            from jolt.plugins import fetch

            class Example(Task):
                requires = ["fetch:alias=zlib,url=https://zlib.net/zlib-1.3.1.tar.gz"]

                def run(self, deps, tools):
                    self.info("The source tree is located at: {fetch[zlib]}")
                    with tools.cwd(self.fetch["zlib"]):
                        tools.run("make")

            """

    name = "fetch"
    alias = Parameter(required=False, help="Name of the task used when referencing content. Defaults to the filename of the fetched file.")
    extract = BooleanParameter(default=True, help="Whether to extract the fetched file.")
    path = Parameter(required=False, help="Destination directory.")
    url = Parameter(help="URL to fetch from.")
    md5 = Parameter(required=False, help="Expected MD5 hash of the fetched file.")
    sha256 = Parameter(required=False, help="Expected SHA256 hash of the fetched file.")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.joltdir = JoltLoader.get().joltdir

        # Set the path to the extraction directory
        if self.path.is_unset():
            self.abspath = self.tools.builddir(utils.canonical(self.short_qualified_name), incremental="always", unique=False)
            if not self._extract():
                # Join with filename if not extracting
                self.abspath = fs.path.join(self.abspath, self._get_filename())
        else:
            self.abspath = fs.path.join(self.joltdir, str(self.path) or self._get_name())

        self.relpath = fs.path.relpath(self.abspath, self.tools.wsroot)

    def _extract(self):
        """ Check if the fetched file should/can be extracted. """
        filename = self._get_filename()

        if not any([filename.endswith(ext) for ext in SUPPORTED_ARCHIVE_TYPES]):
            return False

        return bool(self.extract)

    def _acquire_ws(self):
        # Create the destination directory if it does not exist
        self.tools.rmtree(self.abspath, ignore_errors=True)
        self.tools.mkdir(fs.path.dirname(self.abspath), recursively=True)

        if self._extract():
            with self.tools.tmpdir() as tmpdir, self.tools.cwd(tmpdir):
                filename = self._get_filename()
                self.tools.download(self.url, filename)
                self._verify_sha256(filename)
                self.tools.extract(filename, self.abspath)
        else:
            self.tools.download(self.url, self.abspath)
            self._verify_sha256(self.abspath)

    def acquire(self, artifact, deps, tools, owner):
        self._acquire_ws()
        self._assign_fetch(owner)
        artifact.worktree = fs.path.relpath(self.abspath, owner.joltdir)

    def _assign_fetch(self, task, none=False):
        if not hasattr(task, "fetch"):
            task.fetch = ErrorDict(self)
        if none:
            # None means the git repo is not cloned or checked out
            # and should not be included in the git dictionary
            # of the consuming task yet. If the consuming task
            # requires the git repo for its influence collection,
            # the dict will raise an error. The solution is to
            # assign hash=true to the git requirement which
            # will cause the git repo to be cloned and checked out
            # before the influence collection is performed.
            task.fetch[self._get_name()] = None
        else:
            # Assign the git repo to the consuming task.
            # The git repo is cloned and checked out before
            # any influence collection is performed.
            task.fetch[self._get_name()] = fs.path.relpath(self.abspath, task.joltdir)

    def _get_name(self):
        return str(self.alias) if self.alias.is_set() else self._get_filename()

    def _get_filename(self):
        return fs.path.basename(str(self.url))

    def _verify_sha256(self, filepath):
        if not self.sha256.is_set():
            return
        actual_hash = self.tools.checksum_file(filepath, hashfn=hashlib.sha256)
        expected_hash = str(self.sha256)
        raise_error_if(
            actual_hash != expected_hash,
            f"SHA256 hash mismatch for fetched file '{filepath}': expected {expected_hash}, got {actual_hash}"
        )


TaskRegistry.get().add_task_class(Fetch)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import boost, curl, libedit, libtirpc, lz4, openssl, rapidjson, zlib, zstd
from jolt.plugins import cmake, cxxinfo, git, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_curl")
@attributes.requires("requires_{system}_libedit")
@attributes.requires("requires_{system}_libtirpc")
@attributes.requires("requires_lz4")
@attributes.requires("requires_ncurses")
@attributes.requires("requires_rapidjson")
@attributes.requires("requires_src")
@attributes.requires("requires_ssl")
@attributes.requires("requires_zlib")
@attributes.requires("requires_zstd")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@pkgconfig.requires()
class MySQLBase(cmake.CMake):
    abstract = True
    version = Parameter("9.5.0", help="MySQL version.")
    requires_curl = ["curl"]
    requires_darwin_libedit = ["libedit:shared=true"]
    requires_linux_libedit = requires_darwin_libedit
    requires_linux_libtirpc = ["libtirpc"]
    requires_lz4 = ["lz4"]
    requires_rapidjson = ["rapidjson"]
    requires_src = ["git:url=https://github.com/mysql/mysql-server,path={buildroot}/git-mysql,submodules=true"]
    requires_ssl = ["openssl"]
    requires_zlib = ["zlib"]
    requires_zstd = ["zstd"]
    srcdir = "{git[mysql-server]}"


@cxxinfo.publish(libraries=["mysqlclient"])
class LibMySQLClient(MySQLBase):
    name = "libmysqlclient"
    shared = BooleanParameter(False, help="Build shared libraries")
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "WITH_CURL=system",
        "WITH_EDITLINE=system",
        "WITH_LZ4=system",
        "WITH_RAPIDJSON=system",
        "WITH_SSL=system",
        "WITH_UNIT_TESTS=OFF",
        "WITH_ZLIB=system",
        "WITH_ZSTD=system",
        "WITHOUT_SERVER=ON",
    ]


TaskRegistry.get().add_task_class(LibMySQLClient)



from contextlib import contextmanager, ExitStack, nullcontext
import copy
import hashlib
from os import getenv
from threading import RLock
from collections import OrderedDict
import uuid
import socket
import sys

from jolt import cli
from jolt import common_pb2 as common_pb
from jolt import config
from jolt import log
from jolt import utils
from jolt import colors
from jolt import hooks
from jolt import filesystem as fs
from jolt.error import raise_error_if
from jolt.error import raise_task_error_if
from jolt.influence import HashInfluenceRegistry, TaskRequirementInfluence
from jolt.options import JoltOptions
from jolt.tasks import Alias, Resource, WorkspaceResource


class TaskProxy(object):
    def __init__(self, task, graph, cache, options):
        self.task = task
        self.graph = graph
        self.cache = cache
        self.options = options

        # Direct and transitive dependencies.
        # The dependency chain is broken at
        # selfsustained dependencies that don't
        # require their own dependencies anymore
        # after being executed.
        self.children = []
        self.ancestors = set()

        # All direct and transitive dependencies.
        # Unlike 'children', this set is not filtered
        # from selfsustained tasks.
        self.descendants = set()

        # Unfiltered direct dependencies.
        self.neighbors = []

        self.extensions = []
        self.duration_queued = None
        self.duration_running = None
        self.requirement_aliases = {}

        self._extended_task = None
        self._in_progress = False
        self._completed = False
        self._goal = False
        self._download = True
        self._local = False
        self._network = False
        self._artifacts = []
        self._status = None
        self._finalized = False

        # Consumer task if this is a resource
        self._owner = None

        # List of all artifacts that are produced by this task
        self._artifacts = []

    def __hash__(self):
        return id(self)

    @property
    def artifacts(self):
        return self._artifacts

    def add_artifact(self, artifact):
        self._artifacts.append(artifact)

    def get_artifact(self, name):
        for artifact in self.artifacts:
            if artifact.name == name:
                return artifact
        return None

    @property
    def tools(self):
        return self.task.tools

    @property
    def name(self):
        return self.task.name

    @property
    def canonical_name(self):
        return self.task.canonical_name

    @property
    def qualified_name(self):
        return self.task.qualified_name

    @property
    def short_qualified_name(self):
        return self.task.short_qualified_name

    @property
    def exported_name(self):
        return self.task.exported_name

    @property
    def log_name(self):
        return "({0} {1})".format(self.short_qualified_name, self.identity[:8])

    def log_running_time(self):
        """ Emits an information log line every 5 mins a task has been running. """
        if not self.is_running():
            return
        minutes = int(self.duration_running.seconds / 60)
        if (minutes % 5) == 0 and minutes > 0:
            if minutes >= 60:
                fmt = f"{int(minutes / 60)}h {int(minutes % 60)}min"
            else:
                fmt = f"{int(minutes % 60)}min"
            if self.is_remotely_executed():
                self.info("Remote execution still in progress after {}", fmt)
            else:
                self.info("Execution still in progress after {}", fmt)

    @property
    def identity(self):
        raise_task_error_if(not self._finalized, self, "Task identity read prematurely")

        if self.task.identity is not None:
            return self.task.identity

        # Acquire workspace resources before calculating the identity
        for c in self.children:
            if c.is_workspace_resource():
                c.task.acquire_ws()

        hash = utils.hashfn()
        HashInfluenceRegistry.get().apply_all(self.task, hash)
        self.task.identity = hash.hexdigest()

        return str(self.task.identity)

    @identity.setter
    def identity(self, value):
        self.task.identity = value

    @property
    def instance(self):
        return self.task.instance

    @instance.setter
    def instance(self, value):
        self.task.instance = value

    @property
    def is_unstable(self):
        return self.task.unstable

    @property
    def weight(self):
        return self.task.weight

    @weight.setter
    def weight(self, weight):
        self.task.weight = weight

    def __str__(self):
        return "{0}{1}".format(self.short_qualified_name, "*" if self.is_extension() else '')

    def info(self, fmt, *args, **kwargs):
        self.task.info(fmt + " " + self.log_name, *args, **kwargs)

    def debug(self, fmt, *args, **kwargs):
        log.debug(fmt + " " + self.log_name, *args, **kwargs)

    def verbose(self, fmt, *args, **kwargs):
        log.verbose(fmt + " " + self.log_name, *args, **kwargs)

    def warning(self, fmt, *args, **kwargs):
        self.task.warning(fmt + " " + self.log_name, *args, **kwargs)

    def error(self, fmt, *args, **kwargs):
        self.task.error(fmt + " " + self.log_name, *args, **kwargs)

    def has_children(self):
        return len(self.children) > 0

    def has_ancestors(self):
        return len(self.ancestors) > 0

    def has_artifact(self):
        return self.is_cacheable() and not self.is_alias()

    def has_extensions(self):
        return len(self.extensions) > 0

    def add_extension(self, task):
        if self.is_extension():
            self.get_extended_task().add_extension(task)
        self.extensions.append(task)

    def set_extended_task(self, task):
        self._extended_task = task

    def get_extended_task(self):
        if self.is_extension():
            return self._extended_task.get_extended_task()
        return self

    def deps_available_locally(self):
        for c in self.children:
            if c.is_resource() or c.is_alias():
                continue
            if not c.is_available_locally(persistent_only=True):
                return False
        return True

    def is_alias(self):
        return isinstance(self.task, Alias)

    def is_available_locally(self, extensions=True, persistent_only=True):
        dep_artifacts = []
        if extensions:
            for dep in self.extensions:
                dep_artifacts += dep.artifacts
        artifacts = self._artifacts + dep_artifacts
        if persistent_only:
            artifacts = filter(lambda a: not a.is_session(), artifacts)
        return all(map(self.cache.is_available_locally, artifacts))

    def is_available_remotely(self, extensions=True, cache=True):
        dep_artifacts = []
        if extensions:
            for dep in self.extensions:
                dep_artifacts += dep.artifacts
        artifacts = filter(lambda a: not a.is_session(), self._artifacts + dep_artifacts)
        return all(map(lambda artifact: self.cache.is_available_remotely(artifact, cache=cache), artifacts))

    def is_cacheable(self):
        return self.task.is_cacheable()

    def is_completed(self):
        return self._completed

    def is_downloadable(self):
        return self._download

    def is_extension(self):
        return self._extended_task is not None

    def is_fast(self):
        tasks = [self.task] + [e.task for e in self.extensions]
        return all([task.fast for task in tasks])

    def is_identified(self):
        return self.task.identity is not None

    def is_goal(self, with_extensions=True):
        return self._goal or (with_extensions and any([e.is_goal() for e in self.extensions]))

    def is_local(self):
        if self.is_extension():
            return self.get_extended_task().is_local()
        tasks = [self.task] + [e.task for e in self.extensions]
        return any([task.local for task in tasks])

    def in_progress(self):
        return self._in_progress

    def is_ready(self):
        if self.in_progress():
            return False

        if self.is_extension():
            return False

        return self.graph.is_leaf(self)

    def is_locally_executed(self):
        return self._local

    def is_remotely_executed(self):
        return self._network

    def is_resource(self):
        return isinstance(self.task, Resource)

    def is_workspace_resource(self):
        return isinstance(self.task, WorkspaceResource)

    def is_running(self):
        return self.status() == common_pb.TaskStatus.TASK_RUNNING

    def is_unpackable(self):
        tasks = [self] + self.extensions
        artifacts = []
        for task in tasks:
            artifacts.extend(task._artifacts)
        return any(map(lambda artifact: artifact.is_unpackable(), artifacts))

    def is_unpacked(self):
        tasks = [self] + self.extensions
        artifacts = []
        for task in tasks:
            artifacts.extend(task._artifacts)
        return any(map(lambda artifact: artifact.is_unpacked(), artifacts))

    def is_uploadable(self, artifacts=None):
        tasks = [self] + self.extensions
        if not artifacts:
            artifacts = []
            for task in tasks:
                artifacts.extend(task._artifacts)
        return all(map(lambda artifact: artifact.is_uploadable(), artifacts))

    @contextmanager
    def lock_artifacts(self, discard=False):
        artifacts = []
        stack = ExitStack()
        try:
            for artifact in self.artifacts:
                lock = self.cache.lock_artifact(artifact, discard=discard)
                artifacts.append(stack.enter_context(lock))
            self._artifacts = artifacts
            yield artifacts
        finally:
            stack.close()

    def disable_download(self):
        self._download = False

    def download(self, force=False, session_only=False, persistent_only=False):
        """
        Downloads all artifacts of this task.

        If the task is not downloadable, the method returns True. Failure to
        download persistent artifacts is considered a failure, and the method
        returns False. Session artifacts are not required to be downloaded.

        :param force: Force download even if the artifacts are already available.
        :param session_only: Download only session artifacts.
        :param persistent_only: Download only persistent artifacts.

        """
        if not force and not self.is_downloadable():
            return True
        success = True
        artifacts = self._artifacts
        artifacts_session = list(filter(lambda a: a.is_session(), artifacts))
        artifacts_persistent = list(filter(lambda a: not a.is_session(), artifacts))
        download_all = not session_only and not persistent_only
        if session_only or download_all:
            for artifact in artifacts_session:
                if not self.cache.download(artifact, force=force):
                    self.warning("Failed to download session artifact: {}", artifact.identity)
        if persistent_only or download_all:
            success = all([self.cache.download(artifact, force=force) for artifact in artifacts_persistent])
        return success

    def upload(self, force=False, locked=False, session_only=False, persistent_only=False, artifacts=None):
        artifacts = artifacts or self._artifacts
        if session_only:
            artifacts = list(filter(lambda a: a.is_session(), artifacts))
        if persistent_only:
            artifacts = list(filter(lambda a: not a.is_session(), artifacts))
        if not artifacts:
            return True
        if not self.is_uploadable(artifacts):
            return False
        return all([self.cache.upload(artifact, force=force, locked=locked) for artifact in artifacts])

    def resolve_requirement_alias(self, name):
        return self.requirement_aliases.get(name)

    def set_cancelled(self):
        self.set_status(common_pb.TaskStatus.TASK_CANCELLED)

    def set_passed(self):
        self.set_status(common_pb.TaskStatus.TASK_PASSED)

    def set_failed(self):
        self.set_status(common_pb.TaskStatus.TASK_FAILED)

    def set_skipped(self):
        self.set_status(common_pb.TaskStatus.TASK_SKIPPED)

    def set_downloaded(self):
        self.set_status(common_pb.TaskStatus.TASK_DOWNLOADED)

    def set_uploaded(self):
        self.set_status(common_pb.TaskStatus.TASK_UPLOADED)

    def set_running(self):
        self.set_status(common_pb.TaskStatus.TASK_RUNNING)

    def set_queued(self):
        self.set_status(common_pb.TaskStatus.TASK_QUEUED)

    def status(self):
        return self._status

    def set_status(self, status):
        self._status = status

    def set_in_progress(self):
        self.set_queued()
        self._in_progress = True

    def set_locally_executed(self):
        self._local = True

    def set_remotely_executed(self):
        self._network = True

    def set_goal(self):
        self._goal = True

    def set_owner(self, owner):
        self._owner = owner
        self.task.exported_name = f"{self.short_qualified_name}@@{owner.short_qualified_name}"

    def finalize(self, dag):
        log.debug("Finalizing: " + self.short_qualified_name)

        # Find all direct and transitive dependencies
        self.ancestors = set()
        self.descendants = set()

        self.neighbors = copy.copy(self.children)
        self.neighbors = sorted(self.neighbors, key=lambda n: n.qualified_name)

        for n in self.neighbors:
            self.descendants.add(n)
            self.descendants = self.descendants.union(n.descendants)
            if not n.task.selfsustained:
                self.children.extend(n.children)
            n.ancestors.add(self)

        # Exclude transitive alias and resources dependencies.
        # Workspace resources are included as they may be required by its dependencies.
        self.children = list(
            filter(lambda n: dag.are_neighbors(self, n) or (not n.is_alias() and not n.is_resource()),
                   utils.unique_list(self.children)))

        # Prepare workspace resources for this task so that influence can be calculated
        for child in self.children:
            if not child.is_workspace_resource():
                continue
            child.task.prepare_ws_for(self.task)

        self.descendants = list(self.descendants)

        self.task.influence += [TaskRequirementInfluence(n) for n in self.neighbors]
        self._finalized = True
        self.identity

        hooks.task_created(self)

        return self.identity

    def finalize_artifacts(self):
        self._artifacts.extend(self.task._artifacts(self.cache, self))

    def taint(self, salt=None):
        self.task.taint = salt or uuid.uuid4()
        if salt is None:
            # Only recalculate identity when build is forced, not when salted
            self.identity = None
            self.identity
            # Recreate artifacts
            self._artifacts = []
            self.finalize_artifacts()

        # If this is an alias, taint all children
        if self.is_alias():
            for child in self.children:
                child.taint()

        # Taint all extensions
        for extension in self.extensions:
            extension.taint()

    def queued(self, remote=True):
        self.task.verbose("Task queued " + self.log_name)
        self.duration_queued = utils.duration()
        self.set_queued()
        hooks.task_queued(self)

    def running(self, when=None, what="Execution"):
        if what:
            self.task.info(colors.blue(what + " started " + self.log_name))
            self.set_running()
            hooks.task_started(self)
        self.duration_running = utils.duration() if not when else when

    def running_execution(self, remote=False):
        self.worker = socket.gethostname()
        hooks.task_started_execution(self)
        self.running(what="Remote execution" if remote else "Execution")

    def restarted_execution(self, remote=False):
        hooks.task_finished_execution(self)
        self.task.warning("Remote execution interrupted {}" if remote else "Execution interrupted {}", self.log_name)
        self.queued()

    def started_execution(self, remote=False):
        self.queued()
        self.running_execution(remote=remote)

    def started_download(self):
        self.queued()
        self.running(what="Download")
        hooks.task_started_download(self)

    def started_upload(self):
        self.queued()
        self.running(what="Upload")
        hooks.task_started_upload(self)

    def _failed(self, what="Execution", interrupt=False):
        if interrupt:
            how = "interrupted"
            logfn = self.warning
            self.set_cancelled()
        else:
            how = "failed"
            logfn = self.error
            self.set_failed()

        if self.duration_queued and self.duration_running:
            logfn("{0} {1} after {2} {3}", what, how,
                  self.duration_running,
                  self.duration_queued.diff(self.duration_running))
        elif self.duration_queued:
            logfn("{0} {1} after {2}", what, how, self.duration_queued or utils.duration())
        else:
            logfn("{0} {1} immediately", what, how)

        if self.is_unstable:
            try:
                self.graph.remove_node(self)
            except KeyError:
                self.warning("Pruned task was executed")
            self.graph.add_unstable(self)
            hooks.task_unstable(self)
        else:
            self.graph.add_failed(self)
            hooks.task_failed(self)

    def failed_download(self):
        self._failed("Download")

    def failed_upload(self):
        self._failed("Upload")

    def failed_execution(self, remote=False, interrupt=False):
        self._failed(what="Remote execution" if remote else "Execution", interrupt=interrupt)

    def _finished(self, what="Execution"):
        raise_task_error_if(
            self.is_completed() and not self.is_extension(),
            self, "task has already been completed")
        self._completed = True
        try:
            self.graph.remove_node(self)
        except KeyError:
            self.warning("Pruned task was executed")
        self.task.info(colors.green(what + " finished after {0} {1}" + self.log_name),
                       self.duration_running or "00s",
                       self.duration_queued.diff(self.duration_running))
        hooks.task_finished(self)

    def finished_download(self):
        self.set_downloaded()
        hooks.task_finished_download(self)
        self._finished("Download")

    def finished_upload(self):
        self.set_uploaded()
        hooks.task_finished_upload(self)
        self._finished("Upload")

    def finished_execution(self, remote=False):
        self.set_passed()
        hooks.task_finished_execution(self)
        self._finished(what="Remote execution" if remote else "Execution")

    def skipped(self):
        self._completed = True
        try:
            self.graph.remove_node(self)
        except KeyError:
            pass
        self.set_skipped()
        hooks.task_skipped(self)

    def pruned(self):
        self._completed = True
        try:
            self.graph.remove_node(self)
            self.graph.add_pruned(self)
            hooks.task_pruned(self)
        except KeyError:
            self.warning("Pruned task was already pruned")

    def clean(self, cache, if_expired, onerror=None):
        with self.tools:
            self.task.clean(self.tools)
            for artifact in self.artifacts:
                discarded = cache.discard(artifact, if_expired, onerror=fs.onerror_warning)
                if discarded:
                    log.debug("Discarded: {} ({})", self.short_qualified_name, artifact.identity)
                else:
                    log.debug(" Retained: {} ({})", self.short_qualified_name, artifact.identity)

    def _run_download_dependencies(self, resource_only=False):
        for child in self.children:
            if not child.has_artifact():
                continue

            if child.is_resource() and child.is_local():
                if child.options.worker:
                    # Resource already acquired by the client when running as worker
                    raise_task_error_if(
                        not child.download(force=True),
                        child, "Failed to download task artifact")
                else:
                    # Resource about to be acquired by the client
                    child._run_download_dependencies()
                continue

            if resource_only and not child.is_resource():
                continue

            raise_task_error_if(
                not child.is_completed() and child.is_unstable,
                self, "Task depends on failed task '{}'", child.short_qualified_name)
            if not child.is_available_locally(extensions=False):
                raise_task_error_if(
                    not child.download(force=True),
                    child, "Failed to download task artifact")

    def partially_available_locally(self):
        availability = map(lambda a: self.cache.is_available_locally(a), self.artifacts)
        return any(availability) and not all(availability)

    def _validate_platform(self):
        """ Validates that the task is runnable on the current platform. """
        platform_os, platform_arch = utils.platform_os_arch()

        os = self.task.platform.get("node.os")
        if os:
            os = self.tools.expand(os)
            raise_task_error_if(
                os != platform_os,
                self, f"Task is not runnable on current platform (wants node.os={os})")

        arch = self.task.platform.get("node.arch")
        if arch:
            arch = self.tools.expand(arch)
            raise_task_error_if(
                arch != platform_arch,
                self, f"Task is not runnable on current platform (wants node.arch={arch})")

    def run_acquire(self, artifact, owner, log_prefix=False):
        """
        Acquires a resource and publishes its artifact.

        The artifact is published to the cache even if the acquisition fails.
        """

        try:
            if not self.is_workspace_resource():
                ts = utils.duration()
                log.info(colors.blue("Resource acquisition started ({} for {})"), self.short_qualified_name, owner.short_qualified_name)

            try:
                with log.thread_prefix(owner.identity[:8]) if log_prefix else nullcontext():
                    acquire = getattr(self.task, "acquire_" + artifact.name) if artifact.name != "main" else self.task.acquire
                    acquire(artifact, self.deps, self.tools, owner.task)
            finally:
                # Always commit the resource session artifact to the cache, even if the acquisition failed.
                if not self.is_workspace_resource():
                    self.cache.commit(artifact)

            if not self.is_workspace_resource():
                log.info(colors.green("Resource acquisition finished after {} ({} for {})"), ts, self.short_qualified_name, owner.short_qualified_name)

        except (KeyboardInterrupt, Exception) as e:
            if not self.is_workspace_resource():
                log.error(colors.red("Resource acquisition failed after {} ({} for {})"), ts, self.short_qualified_name, owner.short_qualified_name)
            if self.task.release_on_error:
                with utils.ignore_exception():
                    self.run_release(artifact, owner)
            raise e

    def run_release(self, artifact, owner, log_prefix=False):
        """
        Releases a resource.
        """
        try:
            if not self.is_workspace_resource():
                ts = utils.duration()
                log.info(colors.blue("Resource release started ({} for {})"), self.short_qualified_name, owner.short_qualified_name)

            with log.thread_prefix(owner.identity[:8]) if log_prefix else nullcontext():
                release = getattr(self.task, "release_" + artifact.name) if artifact.name != "main" else self.task.release
                release(artifact, self.deps, self.tools, owner.task)

            if not self.is_workspace_resource():
                log.info(colors.green("Resource release finished after {} ({} for {})"), ts, self.short_qualified_name, owner.short_qualified_name)

        except (KeyboardInterrupt, Exception) as e:
            if not self.is_workspace_resource():
                log.error(colors.red("Resource release failed after {} ({} for {})"), ts, self.short_qualified_name, owner.short_qualified_name)
            raise e

    @contextmanager
    def run_resources(self):
        """
        Acquires and releases resources for the task.

        The method is called by executors before invoking the task proxy's run() method.
        Resource dependencies are acquired and released in reverse order. If an acquisition fails,
        already acquired resources are released in reverse order and the exception is propagated
        to the caller.

        Resource artifacts are always published and uploaded if the acquisition has been started,
        even if the acquisition fails. That way, a failed acquisition can be debugged.
        """
        self._run_download_dependencies(resource_only=True)

        with self._run_resources_no_dep_download():
            yield

    @contextmanager
    def _run_resources_no_dep_download(self):
        # Log messages are prefixed with task identity if resources are acquired in parallel
        log_prefix = False

        # Collect list of resource dependencies
        resource_deps = [child for child in self.children if child.is_resource()]

        if self.options.worker:
            # Exclude local resources when running as worker. They are already acquired by the client.
            resource_deps = [child for child in resource_deps if not child.is_local()]
        elif self.options.network and not self.is_local():
            # Exclude non-local resources in the client when running a network build.
            # They are acquired by the remote worker.
            resource_deps = [child for child in resource_deps if child.is_local()]
            log_prefix = True

        exitstack = ExitStack()
        acquired = []
        try:
            # Acquire resource dependencies in reverse order.
            for resource in reversed(resource_deps):
                # Always discard resource artifacts before acquiring the resource.
                # They should not exist in the cache when the resource is acquired,
                # but may exist if the resource was previously acquired by an interrupted build.
                with resource.lock_artifacts(discard=True) if not resource.is_workspace_resource() else nullcontext():
                    resource.deps = self.cache.get_context(resource)
                    exitstack.enter_context(resource.deps)

                    # Just like tasks, a resource may have multiple artifacts. Run acquire for each artifact.
                    for artifact in resource.artifacts:
                        try:
                            resource.run_acquire(artifact, self, log_prefix=log_prefix)
                            acquired.append(resource)
                        finally:
                            # Always upload the artifact session artifact to the cache, even if the acquisition failed.
                            if not resource.is_workspace_resource():
                                resource.upload(locked=False, session_only=True, artifacts=[artifact])

            yield

        finally:
            for resource in reversed(acquired):
                for artifact in resource.artifacts:
                    resource.run_release(artifact, self, log_prefix=log_prefix)
            exitstack.close()

    def run(self, env, force_upload=False, force_build=False):
        # Download dependency artifacts if not already done
        self._run_download_dependencies()

        with self._run_resources_no_dep_download():
            self._run_task(env, force_upload, force_build)

    def _run_task(self, env, force_upload=False, force_build=False):
        queue = env.queue

        with self.tools:
            available_locally = available_remotely = False

            # Check if task artifact is available locally or remotely,
            # either skip execution or download it if necessary.
            if not force_build:
                available_locally = self.is_available_locally()
                if available_locally and not force_upload:
                    self.skipped()
                    return

                available_remotely = self.cache.download_enabled() and self.is_available_remotely()
                if not available_locally and available_remotely:
                    available_locally = self.download()

                if not available_locally and self.partially_available_locally():
                    force_build = True

            if force_build or not available_locally:
                with log.threadsink() as buildlog:
                    if self.task.is_runnable():
                        log.verbose("Host: {0}", getenv("HOSTNAME", "localhost"))

                    with self.lock_artifacts(discard=not self.is_resource()) as artifacts:
                        exitstack = ExitStack()

                        # Indicates whether session artifacts have been published
                        upload_session_artifacts = False

                        try:
                            context = self.cache.get_context(self)
                            exitstack.enter_context(context)

                            self.running_execution()

                            self._validate_platform()

                            with self.tools.cwd(self.task.joltdir):
                                if self.is_goal() and self.options.debug:
                                    log.info("Entering debug shell")
                                    self.task.debugshell(context, self.tools)

                                try:
                                    # Run task
                                    try:
                                        hooks.task_prerun(self, context, self.tools)
                                        with self.tools.timeout(seconds=config.getint("jolt", "task_timeout")):
                                            self.task.run(context, self.tools)
                                    finally:
                                        hooks.task_postrun(self, context, self.tools)

                                    # Publish persistent artifacts
                                    if not self.is_available_locally(extensions=False):
                                        for artifact in filter(lambda a: not a.is_session(), artifacts):
                                            self.publish(context, artifact, buildlog)
                                    else:
                                        self.info("Publication skipped, already in local cache")

                                finally:
                                    # Session artifacts should be uploaded
                                    upload_session_artifacts = []

                                    # Publish session artifacts to local cache
                                    for artifact in filter(lambda a: a.is_session(), artifacts):
                                        self.publish(context, artifact)
                                        upload_session_artifacts.append(artifact)

                        except KeyboardInterrupt as e:
                            self.failed_execution(interrupt=True)
                            with utils.ignore_exception():
                                exitstack.close()

                            raise e

                        except Exception as e:
                            self.failed_execution(interrupt=queue.is_aborted() if queue else False)

                            with utils.ignore_exception():
                                exitstack.close()

                            if queue is not None and queue.is_aborted():
                                raise KeyboardInterrupt()

                            if cli.debug_enabled:
                                import pdb
                                extype, value, tb = sys.exc_info()
                                pdb.post_mortem(tb)

                            raise e

                        else:
                            self.finished_execution()
                            exitstack.close()

                            # Must upload the artifact while still holding its lock, otherwise the
                            # artifact may become unpack():ed before we have a chance to.
                            if force_upload or force_build or not available_remotely:
                                raise_task_error_if(
                                    not self.upload(force=force_upload, locked=False, persistent_only=True) \
                                    and self.cache.upload_enabled(),
                                    self, "Failed to upload task artifact")

                        finally:
                            # Upload published session artifacts to remote cache
                            raise_task_error_if(
                                upload_session_artifacts \
                                and not self.upload(force=force_upload, locked=False, session_only=True, artifacts=upload_session_artifacts) \
                                and self.cache.upload_enabled(),
                                self, "Failed to upload session artifact")

            elif force_upload or not available_remotely:
                self.started_upload()
                raise_task_error_if(
                    not self.upload(force=force_upload, persistent_only=True) \
                    and self.cache.upload_enabled(),
                    self, "Failed to upload task artifact")
                self.finished_upload()

            else:
                self.skipped()
                self.info("Execution skipped, already in local cache")

            for extension in self.extensions:
                with hooks.task_run(extension):
                    extension.run(env, force_upload, force_build)

    def publish(self, context, artifact, buildlog=None):
        hooks.task_prepublish(self, artifact, self.tools)
        publish = self.task.publish if artifact.is_main() else \
            getattr(self.task, "publish_" + artifact.name)
        publish(artifact, self.tools)
        self.task._verify_influence(context, artifact, self.tools)
        if artifact.is_main() and buildlog:
            with open(fs.path.join(artifact.path, ".build.log"), "w") as f:
                f.write(buildlog.getvalue())
        hooks.task_postpublish(self, artifact, self.tools)
        artifact.get_cache().commit(artifact)

    def raise_for_status(self, log_details=False, log_error=False):
        with self.report() as report:
            report.raise_for_status(log_details, log_error)

    def report(self):
        return self.task.report()

    def unpack(self):
        """ Unpacks all artifacts produced by this task. """
        for artifact in self.artifacts:
            self.cache.unpack(artifact)


class Graph(object):
    def __init__(self):
        self._mutex = RLock()
        self._failed = []
        self._unstable = []
        self._pruned = []
        self._children = OrderedDict()
        self._parents = OrderedDict()

    def add_node(self, node):
        with self._mutex:
            self._children[node] = OrderedDict()
            self._parents[node] = OrderedDict()

    def remove_node(self, node):
        with self._mutex:
            parents = self._parents[node].keys()
            for child in self._children[node]:
                del self._parents[child][node]
            for parent in parents:
                del self._children[parent][node]
            del self._children[node]
            del self._parents[node]

    def add_pruned(self, node):
        self._pruned.append(node)

    def add_edges_from(self, edges):
        with self._mutex:
            for src, dst in edges:
                self._children[src][dst] = None
                self._parents[dst][src] = None

    def successors(self, node):
        return self._children[node].keys()

    def predecessors(self, node):
        return self._parents[node].keys()

    @property
    def topological_nodes(self):
        with self._mutex:
            G = self.clone()
            S = G.roots
            L = []
            while S:
                for n in S:
                    L.append(n)
                    G.remove_node(n)
                S = G.roots
            if len(G.nodes) > 0:
                log.debug("[GRAPH] Graph has cycles between these nodes:")
                for node in G.nodes:
                    log.debug("[GRAPH]   " + node.short_qualified_name)
                raise_error_if(len(G.nodes) > 0, "graph has cycles")
            return L

    def clone(self):
        g = Graph()
        g._children = OrderedDict()
        g._parents = OrderedDict()
        for k, v in self._children.items():
            g._children[k] = copy.copy(v)
        for k, v in self._parents.items():
            g._parents[k] = copy.copy(v)
        return g

    @property
    def artifacts(self):
        artifacts = []
        for node in self.nodes:
            if node.is_cacheable():
                artifacts.extend(node.artifacts)
        return artifacts

    @property
    def persistent_artifacts(self):
        return list(filter(lambda a: not a.is_session(), self.artifacts))

    @property
    def nodes(self):
        with self._mutex:
            return self._children.keys()

    def number_of_tasks(self, filterfn=None):
        return len(list(filter(filterfn, self.nodes)) if filterfn else self.nodes)

    @property
    def failed(self):
        return self._failed

    def add_failed(self, task):
        self._failed.append(task)

    @property
    def unstable(self):
        return self._unstable

    def add_unstable(self, task):
        self._unstable.append(task)

    @property
    def tasks(self):
        with self._mutex:
            return [n for n in self.nodes]

    @property
    def pruned(self):
        return [p for p in self._pruned]

    @property
    def roots(self):
        with self._mutex:
            return [n for n in self.nodes if self.is_root(n)]

    def has_tasks(self):
        with self._mutex:
            return len(self.nodes) > 0

    def get_task(self, qualified_name):
        with self._mutex:
            return self._nodes_by_name.get(qualified_name)

    def get_task_by_identity(self, identity):
        with self._mutex:
            for task in self.nodes:
                if task.identity == identity:
                    return task
        return None

    def select(self, func):
        with self._mutex:
            return [n for n in self.nodes if func(self, n)]

    def debug(self):
        with self._mutex:
            log.debug("[GRAPH] Listing all nodes")
            for node in self.topological_nodes:
                log.debug("[GRAPH]   " + node.short_qualified_name + " ({})", len(self._children[node].keys()))

    def is_leaf(self, node):
        with self._mutex:
            return len(self._children[node].keys()) == 0

    def is_root(self, node):
        with self._mutex:
            return len(self._parents[node].keys()) == 0

    def is_orphan(self, node):
        with self._mutex:
            return self.is_root(node) and self.is_leaf(node)

    def are_neighbors(self, n1, n2):
        with self._mutex:
            return n2 in self._children[n1]


class GraphBuilder(object):
    def __init__(self, registry, cache, options=None, progress=False, buildenv=None):
        self.cache = cache
        self.graph = Graph()
        self.nodes = {}
        self.registry = registry
        self.buildenv = buildenv
        self.progress = progress
        self.options = options or JoltOptions()

    def _get_node(self, progress, name, parent=None):
        name = utils.stable_task_name(name)
        node = self.nodes.get(name)
        if not node:
            task = self.registry.get_task(name, buildenv=self.buildenv)
            node = self.nodes.get(task.qualified_name, None)
            if node is not None:
                return node
            node = TaskProxy(task, self.graph, self.cache, self.options)
            if not node.is_resource() or node.is_workspace_resource():
                self.nodes[node.short_qualified_name] = node
                self.nodes[node.qualified_name] = node
            elif parent:
                node.set_owner(parent)
                if self.buildenv:
                    task._apply_protobuf(self.buildenv)
            if self.options.salt:
                node.taint(self.options.salt)
            self._build_node(progress, node)
            progress.update(1)
        return node

    def _build_node(self, progress, node):
        self.graph.add_node(node)

        if node.task.extends:
            extended_node = self._get_node(progress, node.task.extends, parent=node)
            self.graph.add_edges_from([(node, extended_node)])
            node.set_extended_task(extended_node)
            extended_node.add_extension(node)
            node.children.append(extended_node)
            parent = extended_node.get_extended_task()
        else:
            parent = node

        for requirement in node.task.requires:
            alias, _, task, name = utils.parse_aliased_task_name(requirement)
            child = self._get_node(progress, utils.format_task_name(task, name), parent=node)

            # Create direct edges from alias parents to alias children
            if child.is_alias():
                for child_child in child.children:
                    self.graph.add_edges_from([(parent, child_child)])
                    node.children.append(child_child)
            self.graph.add_edges_from([(parent, child)])
            node.children.append(child)
            if alias:
                node.requirement_aliases[alias] = requirement

        node.children = utils.unique_list(node.children)

        return node

    @contextmanager
    def _progress(self, *args, **kwargs):
        if self.progress:
            with log.progress(*args, **kwargs) as p:
                yield p
        else:
            with log.progress_log(*args, **kwargs) as p:
                yield p

    def build(self, task_list, influence=True):
        with self._progress("Building graph", len(self.graph.tasks), "tasks") as progress:
            goals = [self._get_node(progress, task) for task in task_list]
            self.graph._nodes_by_name = self.nodes

        if influence:
            topological_nodes = self.graph.topological_nodes
            with self._progress("Collecting task influence", len(self.graph.tasks), "tasks") as p:
                for node in reversed(topological_nodes):
                    node.finalize(self.graph)
                    p.update(1)

                # Create artifacts in forward order so that parent identities are available
                # when creating resource artifacts that depend on them.
                for node in topological_nodes:
                    node.finalize_artifacts()

            max_time = 0
            min_time = 0
            for node in topological_nodes:
                max_time += node.task.weight
                node.task.weight += max([a.weight for a in node.ancestors] + [0])
                min_time = max(node.task.weight, min_time)

            log.verbose("Time estimate: {}- {}",
                        utils.duration_diff(min_time),
                        utils.duration_diff(max_time))

        self.graph.requested_goals = goals
        self.graph.goals = []
        for goal in goals:
            goal.set_goal()
            self.graph.goals.append(goal)
            if goal.is_alias():
                for goal_alias in goal.neighbors:
                    goal_alias.set_goal()
                    self.graph.goals.append(goal_alias)

        self.graph.all_nodes = [n for n in self.graph.nodes]

        return self.graph


class PruneStrategy(object):
    def should_prune_requirements(self, task):
        raise NotImplementedError()


class GraphPruner(object):
    def __init__(self, cache, strategy):
        self.cache = cache
        self.strategy = strategy
        self.retained = set()
        self.visited = set()

    def _check_node(self, node):
        if node in self.visited:
            return
        self._progress.update(1)
        self.visited.add(node)
        self.retained.add(node)

        prune = self.strategy.should_prune_requirements(node)
        if not node.task.selfsustained or not prune or node.is_extension():
            utils.map_concurrent(self._check_node, node.neighbors)

    def prune(self, graph):
        with log.progress("Checking availability", 0, " tasks") as p:
            self._progress = p
            for root in graph.roots:
                self._check_node(root)

        for node in graph.goals:
            self.retained.add(node)

        pruned = []
        for node in graph.nodes:
            if node not in self.retained:
                pruned.append(node)
            else:
                log.debug("Retained: {}", node.log_name)
                node.children = [c for c in node.children if c in self.retained]

        for node in pruned:
            log.debug("Excluded: {}", node.log_name)
            node.pruned()

        return graph



from jolt import attributes
from jolt.plugins import autotools, fetch
from jolt.tasks import TaskRegistry


@attributes.common_metadata()
@attributes.requires("requires_src")
class Texinfo(autotools.Autotools):
    name = "texinfo"
    version = "7.2"
    requires_src = ["fetch:alias=src,url=http://ftpmirror.gnu.org/texinfo/texinfo-{version}.tar.xz"]
    srcdir = "{fetch[src]}/texinfo-{version}"


TaskRegistry.get().add_task_class(Texinfo)



from jolt import attributes, Parameter
from jolt.pkgs import cmake, gflags
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_gflags")
@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
class Glog(cmake.CMake):
    name = "glog"
    version = Parameter("0.7.1", help="glog version.")
    options = ["BUILD_SHARED_LIBS=ON"]
    requires_gflags = ["gflags"]
    requires_git = ["git:url=https://github.com/google/glog.git,rev=v{version}"]
    srcdir = "{git[glog]}"
    options = ["CMAKE_CXX_STANDARD=14"]

    def run(self, deps, tools):
        self.warning("GLog is deprecated and unmaintained. Consider using ng-log instead.")
        super().run(deps, tools)


TaskRegistry.get().add_task_class(Glog)



from jolt import attributes, Task


@attributes.common_metadata()
class Rust(Task):
    """ Base class for Rust-based build tasks. """

    abstract = True

    srcdir = None
    """
    Source directory for the Rust project.

    If None, defaults to the task work directory (joltdir).
    """

    def run(self, deps, tools):
        self.builddir = tools.builddir(incremental=True)
        self.installdir = tools.builddir("install")
        with tools.cwd(self.srcdir or self.joltdir):
            tools.run("cargo install --path . --target-dir={builddir} --root={installdir}")

    def publish(self, artifact, tools):
        with tools.cwd(self.installdir):
            artifact.collect("*", symlinks=True)



from jolt.plugins import python
from jolt.tasks import TaskRegistry


@python.requires()
class Python(python.PythonEnv):
    """ Builds and publishes a Python virtual environment with specified packages. """

    name = "python"

    requirements = [
        "setuptools",
        "mako",
        "meson",
        "pyyaml",
    ]


TaskRegistry.get().add_task_class(Python)



from jolt import attributes, Parameter
from jolt.pkgs import cmake
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
class Re2c(cmake.CMake):
    name = "re2c"
    version = Parameter("4.4", help="re2c version.")
    requires_git = ["git:url=https://github.com/skvadrik/re2c.git,rev={version}"]
    srcdir = "{git[re2c]}"


TaskRegistry.get().add_task_class(Re2c)



from jolt import attributes, Alias, Download, Parameter
from jolt.tasks import TaskRegistry


@attributes.arch
@attributes.system
@attributes.common_metadata()
class CMakeBin(Download):
    name = "cmake/bin"
    version = Parameter("4.2.1", help="CMake version.")
    url = ["https://github.com/Kitware/CMake/releases/download/v{version}/cmake-{version}-{cmake_system}-{cmake_arch}{cmake_ext}"]
    collect = [{"files": "*", "cwd": "cmake-{version}-{cmake_system}-{cmake_arch}{cmake_contents}"}]

    @property
    def cmake_arch(self):
        if self.system == "darwin":
            return "universal"
        if self.arch == "amd64":
            return "x86_64"
        return self.arch

    @property
    def cmake_ext(self):
        if self.system == "windows":
            return ".zip"
        return ".tar.gz"

    @property
    def cmake_system(self):
        if self.system == "darwin":
            return "macos"
        return self.system

    @property
    def cmake_contents(self):
        if self.system == "darwin":
            return "/CMake.app/Contents"
        return ""


class CMake(Alias):
    name = "cmake"
    version = Parameter("4.2.1", help="CMake version.")
    requires = ["cmake/bin:version={version}"]


TaskRegistry.get().add_task_class(CMake)
TaskRegistry.get().add_task_class(CMakeBin)



from jolt import attributes, BooleanParameter, Parameter
from jolt.plugins import cmake, cxxinfo, git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_src")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["ryml", "c4core"])
class RapidYAML(cmake.CMake):
    name = "rapidyaml"
    version = Parameter("0.7.0", help="RapidYAML version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_src = ["git:url=https://github.com/biojppm/rapidyaml.git,rev=v{version},submodules=true"]
    srcdir = "{git[rapidyaml]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "RYML_BUILD_TESTS=OFF",
    ]


TaskRegistry.get().add_task_class(RapidYAML)



from jolt import attributes, BooleanParameter, Parameter, Task
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["gmock", "gtest"])
class GoogleTest(cmake.CMake):
    name = "google/test"
    version = Parameter("1.12.1", help="GoogleTest version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/google/googletest.git,rev=release-{version}"]
    srcdir = "{git[googletest]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
    ]


class GTestMain(Task):
    name = "google/test/main"
    extends = "google/test"

    def publish(self, artifact, tools):
        artifact.cxxinfo.libraries.append("gtest_main")


class GmockMain(Task):
    name = "google/mock/main"
    extends = "google/test"

    def publish(self, artifact, tools):
        artifact.cxxinfo.libraries.append("gmock_main")


TaskRegistry.get().add_task_class(GoogleTest)
TaskRegistry.get().add_task_class(GTestMain)
TaskRegistry.get().add_task_class(GmockMain)



from jolt import attributes, Parameter, BooleanParameter
from jolt.pkgs import libpciaccess, meson
from jolt.plugins import git, meson
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_{intel[intel,no_intel]}")
@meson.requires()
class Libdrm(meson.Meson):
    name = "libdrm"
    version = Parameter("2.4.131", help="Libdrm version.")

    amdgpu = BooleanParameter(True, help="Enable AMD GPU support.")
    etnaviv = BooleanParameter(True, help="Enable Etnaviv support.")
    exynos = BooleanParameter(True, help="Enable Exynos support.")
    freedreno = BooleanParameter(True, help="Enable Freedreno support.")
    intel = BooleanParameter(True, help="Enable Intel support.")
    nouveau = BooleanParameter(True, help="Enable Nouveau support.")
    omap = BooleanParameter(True, help="Enable OMAP support.")
    radeon = BooleanParameter(True, help="Enable Radeon support.")
    tegra = BooleanParameter(True, help="Enable Tegra support.")
    vc4 = BooleanParameter(True, help="Enable VC4 support.")
    vmwgfx = BooleanParameter(True, help="Enable VMware graphics support.")

    requires_git = ["git:url=https://gitlab.freedesktop.org/mesa/drm.git,rev=libdrm-{version}"]
    requires_intel = ["libpciaccess"]
    srcdir = "{git[drm]}"
    options = [
        "amdgpu={amdgpu[enabled,disabled]}",
        "etnaviv={etnaviv[enabled,disabled]}",
        "exynos={exynos[enabled,disabled]}",
        "freedreno={freedreno[enabled,disabled]}",
        "intel={intel[enabled,disabled]}",
        "nouveau={nouveau[enabled,disabled]}",
        "omap={omap[enabled,disabled]}",
        "radeon={radeon[enabled,disabled]}",
        "tegra={tegra[enabled,disabled]}",
        "vc4={vc4[enabled,disabled]}",
        "vmwgfx={vmwgfx[enabled,disabled]}",
    ]


TaskRegistry.get().add_task_class(Libdrm)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake, nlohmann_json
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_json")
@cmake.requires()
@cmake.use_ninja()
class Inja(cmake.CMake):
    name = "inja"
    version = Parameter("3.5.0", help="Inja version.")
    tests = BooleanParameter(False, help="Build tests.")
    requires_git = ["git:url=https://github.com/pantor/inja.git,rev=v{version}"]
    requires_json = ["nlohmann/json"]
    srcdir = "{git[inja]}"
    options = [
        "BUILD_TESTING={tests[ON,OFF]}",
        "INJA_USE_EMBEDDED_JSON=OFF",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.cxxinfo.incpaths.append("include")


TaskRegistry.get().add_task_class(Inja)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["yaml-cpp"])
class YamlCPP(cmake.CMake):
    name = "yaml-cpp"
    version = Parameter("bbf8bdb", help="yaml-cpp version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/jbeder/yaml-cpp.git,rev={version}"]
    srcdir = "{git[yaml-cpp]}"
    options = [
        "CMAKE_POLICY_VERSION_MINIMUM=3.5",
        "YAML_BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "YAML_BUILD_TOOLS=OFF",
    ]


TaskRegistry.get().add_task_class(YamlCPP)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake, libogg
from jolt.plugins import cxxinfo, git, cmake, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_ogg")
@cmake.requires()
@cmake.use_ninja()
@pkgconfig.requires()
@cxxinfo.publish(libraries=["vorbis", "vorbisenc", "vorbisfile"])
class Libvorbis(cmake.CMake):
    name = "libvorbis"
    version = Parameter("1.3.7", help="libvorbis version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/xiph/vorbis.git,rev=v{version}"]
    requires_ogg = ["libogg"]
    srcdir = "{git[vorbis]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "CMAKE_POLICY_VERSION_MINIMUM=3.5",
    ]


TaskRegistry.get().add_task_class(Libvorbis)



from jolt import attributes, Parameter, Task
from jolt.pkgs import cbindgen, glslang, libdrm, libglvnd, libva, x11
from jolt.pkgs import meson, rust, spirv_llvm, wayland
from jolt.plugins import git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_cbindgen")
@attributes.requires("requires_git")
@attributes.requires("requires_glslang")
@attributes.requires("requires_libdrm")
@attributes.requires("requires_llvm")
@attributes.requires("requires_libglvnd")
@attributes.requires("requires_libva")
@attributes.requires("requires_libx11_xcb")
@attributes.requires("requires_libxshmfence")
@attributes.requires("requires_libxxf86vm")
@attributes.requires("requires_meson")
@attributes.requires("requires_rust")
@attributes.requires("requires_spirv_llvm")
@attributes.requires("requires_wayland")
@attributes.requires("requires_xrandr")
@attributes.common_metadata()
class Mesa(Task):
    name = "mesa"
    version = Parameter("25.3.2", help="Mesa version.")

    requires_cbindgen = ["cbindgen"]
    requires_git = ["git:url=https://gitlab.freedesktop.org/mesa/mesa.git,rev=mesa-{version}"]
    requires_glslang = ["glslang"]
    requires_libdrm = ["libdrm"]
    requires_libglvnd = ["libglvnd"]
    requires_libx11_xcb = ["libx11-xcb"]
    requires_libxshmfence = ["libxshmfence"]
    requires_libxxf86vm = ["libxxf86vm"]
    requires_llvm = ["clang", "llvm"]
    requires_libva = ["libva"]
    requires_meson = ["meson"]
    requires_rust = ["rust", "rust-bindgen"]
    requires_spirv_llvm = ["spirv-llvm-translator"]
    requires_wayland = ["wayland"]
    requires_xrandr = ["libxrandr"]

    def run(self, deps, tools):
        try:
            # Patch broken meson.build
            with tools.cwd("{git[mesa]}"):
                tools.replace_in_file(
                    "src/gbm/backends/dri/meson.build",
                    "deps_gbm_dri = []",
                    "deps_gbm_dri = [dep_xcb_dri3]"
                )

            ac = tools.meson()
            ac.configure(
                "{git[mesa]}",
                "buildtype=release",
                "gallium-drivers=auto",
                "glvnd=true",
                "glx=dri",
                "platforms=x11,wayland",
                "shared-llvm=disabled",
                "vulkan-drivers=auto",
            )
            ac.build()
            ac.install()
        finally:
            # Revert patch
            with tools.cwd("{git[mesa]}"):
                tools.replace_in_file(
                    "src/gbm/backends/dri/meson.build",
                    "deps_gbm_dri = [dep_xcb_dri3]",
                    "deps_gbm_dri = []"
                )

    def publish(self, artifact, tools):
        ac = tools.meson()
        ac.publish(artifact)


TaskRegistry.get().add_task_class(Mesa)



from jolt import attributes, BooleanParameter, Parameter
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cmake.options("options_pic_{pic[on,off]}")
@cxxinfo.publish(libraries=["lzma"])
class Xz(cmake.CMake):
    name = "xz"
    version = Parameter("5.8.2", help="xz version.")
    pic = BooleanParameter(True, help="Build with position independent code.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/tukaani-project/xz.git,rev=v{version}"]
    srcdir = "{git[xz]}"
    options_pic_on = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "CMAKE_POSITION_INDEPENDENT_CODE=ON",
    ]


TaskRegistry.get().add_task_class(Xz)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["flatbuffers"])
class Flatbuffers(cmake.CMake):
    name = "flatbuffers"
    version = Parameter("25.9.23", help="Flatbuffers version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/google/flatbuffers.git,rev=v{version}"]
    srcdir = "{git[flatbuffers]}"
    options = [
        "FLATBUFFERS_BUILD_FLATLIB={shared[OFF,ON]}",
        "FLATBUFFERS_BUILD_SHAREDLIB={shared[ON,OFF]}",
        "FLATBUFFERS_BUILD_TESTS=OFF",
    ]


TaskRegistry.get().add_task_class(Flatbuffers)



from jolt import attributes, Parameter
from jolt.pkgs import meson, zlib
from jolt.plugins import git, meson
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_zlib")
@meson.requires()
class GLib(meson.Meson):
    name = "glib"
    version = Parameter("2.86.3", help="glib version.")
    requires_git = ["git:url=https://gitlab.gnome.org/GNOME/glib.git,rev={version},submodules=true"]
    requires_zlib = ["zlib"]
    srcdir = "{git[glib]}"


TaskRegistry.get().add_task_class(GLib)



from jolt import attributes
from jolt import Download
from jolt import Parameter
from jolt.tasks import TaskRegistry


@attributes.attribute("collect", "collect_{system}")
@attributes.attribute("url", "url_{system}")
@attributes.system
class OpenJDK(Download):
    name = "openjdk"
    version = Parameter("25.0.1", const=True, help="OpenJDK version.")

    collect_darwin = [{"files": "*", "cwd": f"jdk-{version}.jdk/Contents/Home"}]
    collect_linux = [{"files": "*", "cwd": f"jdk-{version}"}]
    collect_windows = collect_linux
    url_darwin = "https://download.java.net/java/GA/jdk25.0.1/2fbf10d8c78e40bd87641c434705079d/8/GPL/openjdk-25.0.1_macos-x64_bin.tar.gz"
    url_linux = "https://download.java.net/java/GA/jdk25.0.1/2fbf10d8c78e40bd87641c434705079d/8/GPL/openjdk-25.0.1_linux-x64_bin.tar.gz"
    url_windows = "https://download.java.net/java/GA/jdk25.0.1/2fbf10d8c78e40bd87641c434705079d/8/GPL/openjdk-25.0.1_windows-x64_bin.zip"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.PATH.append("bin")


TaskRegistry.get().add_task_class(OpenJDK)



from jolt import attributes, Parameter
from jolt.plugins import cmake, fetch
from jolt.tasks import TaskRegistry


@attributes.requires("requires_src")
@cmake.requires()
@cmake.use_ninja()
class CLI11(cmake.CMake):
    name = "cli11"
    version = Parameter("2.6.1", help="CLI11 version.")
    requires_src = ["fetch:alias=src,url=https://github.com/CLIUtils/CLI11/archive/refs/tags/v{version}.tar.gz"]
    srcdir = "{fetch[src]}/CLI11-{version}"
    options = [
        "CLI11_BUILD_EXAMPLES=OFF",
        "CLI11_BUILD_TESTS=OFF",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.cxxinfo.incpaths.append("include")


TaskRegistry.get().add_task_class(CLI11)



from jolt import attributes, influence, Parameter, Task
from jolt.plugins import git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
class BusyBox(Task):
    name = "busybox"
    defconfig = Parameter(required=False, help="BusyBox defconfig file.")
    version = Parameter("1.37.0", help="BusyBox version.")
    requires_git = ["git:url=git://busybox.net/busybox.git,rev={version_tag}"]
    srcdir = "{git[busybox]}"

    @property
    def version_tag(self):
        return str(self.version).replace(".", "_")

    @property
    def defconfig_arg(self):
        if self.defconfig:
            return f"{self.defconfig}_defconfig"
        return "defconfig"

    def run(self, deps, tools):
        self.outdir = tools.builddir()
        self.installdir = tools.builddir("install")

        with tools.cwd(self.srcdir):
            tools.run("make O={outdir} {defconfig_arg}")
            tools.run("make O={outdir} -j{}", tools.cpu_count())
            tools.run("make O={outdir} install DESTDIR={installdir}")

    def publish(self, artifact, tools):
        artifact.environ.PATH.append("bin")
        with tools.cwd(self.installdir):
            artifact.collect("*", symlinks=True)


TaskRegistry.get().add_task_class(BusyBox)



from jolt import attributes, Alias, BooleanParameter, Parameter
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class Zlib(cmake.CMake):
    name = "zlib"
    version = Parameter("1.3.1", help="Zlib version.")
    pic = BooleanParameter(False, "Build with position independent code.")
    shared = BooleanParameter(False, "Enable shared libraries.")
    requires_git = ["git:clean=true,url=https://github.com/madler/zlib.git,rev=v{version}"]
    srcdir = "{git[zlib]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "CMAKE_POSITION_INDEPENDENT_CODE={pic[ON,OFF]}",
        "ZLIB_BUILD_EXAMPLES=OFF",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.CMAKE_PREFIX_PATH.append(".")
        if self.system == "windows":
            self.publish_windows(artifact, tools)
        else:
            self.publish_unix(artifact, tools)

    def publish_unix(self, artifact, tools):
        artifact.cxxinfo.libraries.append("z")
        with tools.cwd(artifact.path):
            if self.shared:
                for lib in tools.glob("lib*/*.a"):
                    tools.unlink(lib)
            else:
                for lib in tools.glob("lib*/*.so*") + tools.glob("lib*/*.dylib*"):
                    tools.unlink(lib)

    def publish_windows(self, artifact, tools):
        with tools.cwd(artifact.path):
            if self.shared:
                artifact.cxxinfo.libraries.append("zlib")
                for lib in tools.glob("lib*/zlibstatic.*"):
                    tools.unlink(lib)
            else:
                artifact.cxxinfo.libraries.append("zlibstatic")
                for lib in tools.glob("lib*/zlib.*"):
                    tools.unlink(lib)


@attributes.requires("requires_git")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class ZlibNg(cmake.CMake):
    name = "zlib-ng"
    version = Parameter("2.3.2", help="Zlib version.")
    pic = BooleanParameter(False, "Build with position independent code.")
    shared = BooleanParameter(False, "Enable shared libraries.")
    requires_git = ["git:url=https://github.com/zlib-ng/zlib-ng.git,rev={version}"]
    srcdir = "{git[zlib-ng]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "BUILD_TESTING=OFF",
        "CMAKE_POSITION_INDEPENDENT_CODE={pic[ON,OFF]}",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.CMAKE_PREFIX_PATH.append(".")
        if self.system == "windows":
            artifact.cxxinfo.libraries.append("zlibstatic-ng")
        else:
            artifact.cxxinfo.libraries.append("z-ng")


class VirtualZlib(Alias):
    name = "virtual/zlib"
    pic = BooleanParameter(False, "Build with position independent code.")
    shared = BooleanParameter(False, "Enable shared libraries.")
    requires = ["zlib:pic={pic},shared={shared}"]


TaskRegistry.get().add_task_class(Zlib)
TaskRegistry.get().add_task_class(ZlibNg)
TaskRegistry.get().add_task_class(VirtualZlib)



from jolt import attributes, Parameter
from jolt.pkgs import meson
from jolt.plugins import git, meson
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_zlib")
@meson.requires()
class Libpciaccess(meson.Meson):
    name = "libpciaccess"
    version = Parameter("0.18.1", help="pciaccess version.")
    requires_git = ["git:url=https://gitlab.freedesktop.org/xorg/lib/libpciaccess.git,rev=libpciaccess-{version}"]
    requires_zlib = ["zlib"]
    srcdir = "{git[libpciaccess]}"


TaskRegistry.get().add_task_class(Libpciaccess)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_abseil")
@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["re2"])
class RE2(cmake.CMake):
    name = "re2"
    version = Parameter("2025-11-05", help="re2 version.")
    pic = BooleanParameter(False, help="Build with position independent code.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_abseil = ["abseil:shared={shared}"]
    requires_git = ["git:url=https://github.com/google/re2.git,rev={version}"]
    srcdir = "{git[re2]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "CMAKE_POSITION_INDEPENDENT_CODE={pic[ON,OFF]}",
        "RE2_BUILD_TESTING=OFF",
    ]


TaskRegistry.get().add_task_class(RE2)



from jolt import attributes, BooleanParameter, Parameter
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
class Brotli(cmake.CMake):
    name = "brotli"
    version = Parameter("1.2.0", help="Brotli version.")
    pic = BooleanParameter(False, help="Build with position independent code.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/google/brotli.git,rev=v{version},submodules=true"]
    srcdir = "{git[brotli]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "CMAKE_POSITION_INDEPENDENT_CODE={pic[ON,OFF]}",
    ]


TaskRegistry.get().add_task_class(Brotli)



from jolt import attributes, Parameter
from jolt.pkgs import libexpat, libffi, libxml2
from jolt.plugins import git, meson
from jolt.tasks import TaskRegistry


@attributes.requires("requires_expat")
@attributes.requires("requires_git")
@attributes.requires("requires_libffi")
@attributes.requires("requires_libxml2")
@meson.requires()
class Wayland(meson.Meson):
    name = "wayland"
    version = Parameter("1.24.0", help="Wayland version.")
    requires_expat = ["libexpat"]
    requires_git = ["git:url=https://gitlab.freedesktop.org/wayland/wayland.git,rev={version}"]
    requires_libffi = ["libffi"]
    requires_libxml2 = ["libxml2"]
    srcdir = "{git[wayland]}"
    options = [
        "documentation=false",
        "tests=false",
    ]


TaskRegistry.get().add_task_class(Wayland)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import openssl
from jolt.plugins import cmake, cxxinfo, git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_ssl_{ssl[on,off]}")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class PahoMQTT_C(cmake.CMake):
    name = "paho/mqtt-c"
    version = Parameter("1.3.15", help="Paho MQTT C version.")
    shared = BooleanParameter(False, help="Build shared libraries")
    ssl = BooleanParameter(True, help="Enable SSL support")
    requires_git = ["git:url=https://github.com/eclipse/paho.mqtt.c.git,rev=v{version}"]
    requires_ssl_on = ["openssl"]
    srcdir = "{git[paho.mqtt.c]}"
    options = [
        "PAHO_BUILD_SHARED={shared[ON,OFF]}",
        "PAHO_BUILD_STATIC={shared[OFF,ON]}",
        "PAHO_WITH_SSL={ssl[ON,OFF]}",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.shared:
            artifact.cxxinfo.libraries.append("paho-mqtt3c")
        else:
            artifact.cxxinfo.libraries.append("paho-mqtt3c-static")


@attributes.requires("requires_git")
@attributes.requires("requires_mqtt_c")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class PahoMQTT_CXX(cmake.CMake):
    name = "paho/mqtt-cxx"
    version = Parameter("1.5.2", help="Paho MQTT C++ version.")
    shared = BooleanParameter(False, help="Build shared libraries")
    ssl = BooleanParameter(True, help="Enable SSL support")
    requires_git = ["git:url=https://github.com/eclipse/paho.mqtt.cpp.git,rev=v{version}"]
    requires_mqtt_c = ["paho/mqtt-c:shared={shared},ssl={ssl}"]
    srcdir = "{git[paho.mqtt.cpp]}"
    options = [
        "PAHO_BUILD_SHARED={shared[ON,OFF]}",
        "PAHO_BUILD_STATIC={shared[OFF,ON]}",
        "PAHO_WITH_SSL={ssl[ON,OFF]}",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.shared:
            artifact.cxxinfo.libraries.append("paho-mqttpp3")
        else:
            artifact.cxxinfo.libraries.append("paho-mqttpp3-static")


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class PahoMQTT_Embedded_C(cmake.CMake):
    name = "paho/mqtt-embedded-c"
    version = "1.1.0"
    requires_git = ["git:url=https://github.com/eclipse/paho.mqtt.embedded-c.git,rev=v{version}"]
    srcdir = "{git[paho.mqtt.embedded-c]}"
    options = ["CMAKE_POLICY_VERSION_MINIMUM=3.5"]


TaskRegistry.get().add_task_class(PahoMQTT_C)
TaskRegistry.get().add_task_class(PahoMQTT_CXX)
TaskRegistry.get().add_task_class(PahoMQTT_Embedded_C)



from jolt import attributes


class Libtool(object):
    def __init__(self, tools):
        self.tools = tools

    def relocate(self, artifact, dirs=["lib"], prefix=None):
        prefix = str(artifact.strings.install_prefix) if prefix is None else prefix
        for dir in dirs:
            with self.tools.cwd(artifact.path, dir):
                for file in self.tools.glob("*.la"):
                    self.tools.replace_in_file(file, prefix, artifact.final_path)


def relocate(dirs=["lib"]):
    """
    Relocate libtool archive files (.la) published by task.

    When an artifact is published, all .la files found in the specified
    directories will have their install prefix updated to the artifact's final path.
    The original install prefix is stored in the artifact's strings metadata
    under the key 'libtool_prefix' for use during unpacking.

      :param dirs: List of directories inside the artifact to relocate .la files in.

    """

    def decorate(cls):
        original_publish = cls.publish
        original_unpack = cls.unpack

        def publish(self, artifact, tools):
            original_publish(self, artifact, tools)
            lt = Libtool(tools)
            lt.relocate(artifact, dirs, prefix=artifact.strings.install_prefix)
            artifact.libtool_prefix = artifact.final_path

        def unpack(self, artifact, tools):
            original_unpack(self, artifact, tools)
            lt = Libtool(tools)
            lt.relocate(artifact, dirs, prefix=artifact.strings.libtool_prefix)
            artifact.strings.libtool_prefix = None

        cls.publish = publish
        cls.unpack = unpack

        return cls

    return decorate


def requires():
    """ Decorator to add Libtool requirement to a task. """

    import jolt.pkgs.libtool

    def decorate(cls):
        cls = attributes.requires("requires_libtool")(cls)
        cls.requires_libtool = ["libtool"]
        return cls

    return decorate



from jolt import Alias, BooleanParameter
from jolt.pkgs import boringssl, openssl
from jolt.tasks import TaskRegistry


class VirtualSsl(Alias):
    name = "virtual/ssl"
    pic = BooleanParameter(False, "Build with position independent code.")
    shared = BooleanParameter(False, help="Use shared libraries.")
    requires = ["openssl:shared={shared}"]


TaskRegistry.get().add_task_class(VirtualSsl)



from jolt import attributes, Parameter, Task
from jolt.plugins import fetch, autotools, libtool, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@autotools.requires()
class XorgMacros(autotools.Autotools):
    name = "xorg/macros"
    version = Parameter("1.20.2", help="Xorg Macros version.")

    requires_git = ["git:url=https://gitlab.freedesktop.org/xorg/util/macros.git,rev=util-macros-{version}"]
    srcdir = "{git[macros]}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@autotools.requires()
class Inputproto(autotools.Autotools):
    name = "xorg/inputproto"
    version = Parameter("2.3.2", help="Inputproto version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/proto/inputproto-{version}.tar.gz"]
    srcdir = "{fetch[src]}/inputproto-{version}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@autotools.requires()
class Glproto(autotools.Autotools):
    name = "xorg/glproto"
    version = Parameter("1.4.17", help="Glproto version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/proto/glproto-{version}.tar.gz"]
    srcdir = "{fetch[src]}/glproto-{version}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@autotools.requires()
class Kbproto(autotools.Autotools):
    name = "xorg/kbproto"
    version = Parameter("1.0.7", help="Kbproto version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/proto/kbproto-{version}.tar.gz"]
    srcdir = "{fetch[src]}/kbproto-{version}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@autotools.requires()
class Randrproto(autotools.Autotools):
    name = "xorg/randrproto"
    version = Parameter("1.5.0", help="Randrproto version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/proto/randrproto-{version}.tar.gz"]
    srcdir = "{fetch[src]}/randrproto-{version}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@autotools.requires()
class Renderproto(autotools.Autotools):
    name = "xorg/renderproto"
    version = Parameter("0.11.1", help="Renderproto version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/proto/renderproto-{version}.tar.gz"]
    srcdir = "{fetch[src]}/renderproto-{version}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@attributes.requires("requires_xcb")
@autotools.requires()
@pkgconfig.requires()
class XcbKeysyms(autotools.Autotools):
    name = "xorg/xcb-keysyms"
    version = Parameter("0.4.1", help="XCB Keysyms version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/lib/xcb-util-keysyms-{version}.tar.xz"]
    requires_xcb = ["xorg/libxcb"]
    srcdir = "{fetch[src]}/xcb-util-keysyms-{version}"


@attributes.requires("requires_git")
@attributes.requires("requires_xorg_macros")
@autotools.requires()
@pkgconfig.requires()
class XcbProto(autotools.Autotools):
    name = "xorg/xcb-proto"
    version = Parameter("1.17.0", help="Libxcb_proto version.")
    requires_git = ["git:url=https://gitlab.freedesktop.org/xorg/proto/xcbproto.git,rev=xcb-proto-{version}"]
    requires_xorg_macros = ["xorg/macros"]
    srcdir = "{git[xcbproto]}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@attributes.requires("requires_xproto")
@autotools.requires()
@libtool.relocate()
@pkgconfig.requires()
class Xau(autotools.Autotools):
    name = "xorg/libxau"
    version = Parameter("1.0.12", help="Xau version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/lib/libXau-{version}.tar.gz"]
    requires_xproto = ["xorg/xproto"]
    srcdir = "{fetch[src]}/libXau-{version}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@attributes.requires("requires_xproto")
@autotools.requires()
@libtool.relocate()
@pkgconfig.requires()
class LibXdmcp(autotools.Autotools):
    name = "xorg/libxdmcp"
    version = Parameter("1.1.5", help="LibXdmcp version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/lib/libXdmcp-{version}.tar.gz"]
    requires_xproto = ["xorg/xproto"]
    srcdir = "{fetch[src]}/libXdmcp-{version}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@attributes.requires("requires_x11")
@attributes.requires("requires_xextproto")
@attributes.requires("requires_xproto")
@autotools.requires()
@pkgconfig.requires()
class Xext(autotools.Autotools):
    name = "xorg/libxext"
    version = Parameter("1.3.6", help="Xext version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/lib/libXext-{version}.tar.gz"]
    requires_x11 = ["xorg/libx11"]
    requires_xextproto = ["xorg/xextproto"]
    requires_xproto = ["xorg/xproto"]
    srcdir = "{fetch[src]}/libXext-{version}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@autotools.requires()
class Xextproto(autotools.Autotools):
    name = "xorg/xextproto"
    version = Parameter("7.3.0", help="Xextproto version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/proto/xextproto-{version}.tar.gz"]
    srcdir = "{fetch[src]}/xextproto-{version}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@autotools.requires()
class Xproto(autotools.Autotools):
    name = "xorg/xproto"
    version = Parameter("7.0.31", help="Xproto version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/proto/xproto-{version}.tar.gz"]
    srcdir = "{fetch[src]}/xproto-{version}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@autotools.requires()
class Xf86vidmodeproto(autotools.Autotools):
    name = "xorg/xf86vidmodeproto"
    version = Parameter("2.3.1", help="Xf86vidmodeproto version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/proto/xf86vidmodeproto-{version}.tar.gz"]
    srcdir = "{fetch[src]}/xf86vidmodeproto-{version}"


@attributes.requires("requires_macros")
@attributes.requires("requires_src")
@autotools.requires()
class Xtrans(autotools.Autotools):
    name = "xorg/xtrans"
    version = Parameter("1.6.0", help="Xtrans version.")
    requires_macros = ["xorg/macros"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/lib/xtrans-{version}.tar.gz"]
    srcdir = "{fetch[src]}/xtrans-{version}"


@attributes.requires("requires_randrproto")
@attributes.requires("requires_src")
@attributes.requires("requires_xext")
@attributes.requires("requires_xrender")
@autotools.requires()
@libtool.relocate()
class Xrandr(autotools.Autotools):
    name = "xorg/libxrandr"
    version = Parameter("1.5.4", help="Xrandr version.")
    requires_randrproto = ["xorg/randrproto"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/lib/libXrandr-{version}.tar.gz"]
    requires_xext = ["xorg/libxext"]
    requires_xrender = ["xorg/libxrender"]
    srcdir = "{fetch[src]}/libXrandr-{version}"


@attributes.requires("requires_macros")
@attributes.requires("requires_renderproto")
@attributes.requires("requires_src")
@attributes.requires("requires_x11")
@autotools.requires()
@libtool.relocate()
@pkgconfig.requires()
class Xrender(autotools.Autotools):
    name = "xorg/libxrender"
    version = Parameter("0.9.12", help="Xrender version.")
    requiers_macros = ["xorg/macros"]
    requires_renderproto = ["xorg/renderproto"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/lib/libXrender-{version}.tar.gz"]
    requires_x11 = ["xorg/libx11"]
    srcdir = "{fetch[src]}/libXrender-{version}"


@attributes.requires("requires_git")
@attributes.requires("requires_xproto")
@attributes.requires("requires_xorg_macros")
@pkgconfig.to_cxxinfo(["xshmfence"])
@pkgconfig.requires()
@autotools.requires()
@libtool.relocate()
@libtool.relocate()
class Libxshmfence(autotools.Autotools):
    name = "xorg/libxshmfence"
    version = Parameter("1.3.3", help="Libxshmfence version.")

    requires_git = ["fetch:alias=src,url=https://www.x.org/releases/individual/lib/libxshmfence-{version}.tar.gz"]
    requires_xproto = ["xproto"]
    requires_xorg_macros = ["xorg/macros"]
    srcdir = "{fetch[src]}/libxshmfence-{version}"


@attributes.common_metadata()
@attributes.requires("requires_libx11")
@attributes.requires("requires_src")
@attributes.requires("requires_xext")
@attributes.requires("requires_xextproto")
@attributes.requires("requires_xf86vidmodeproto")
@attributes.requires("requires_xorg_macros")
@attributes.requires("requires_xproto")
@autotools.requires()
@pkgconfig.requires()
@libtool.relocate()
class Libxxf86vm(autotools.Autotools):
    name = "xorg/libxxf86vm"
    version = "1.1.6"
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/lib/libXxf86vm-{version}.tar.gz"]
    srcdir = "{fetch[src]}/libXxf86vm-{version}"
    requires_libx11 = ["xorg/libx11"]
    requires_xext = ["xorg/libxext"]
    requires_xextproto = ["xorg/xextproto"]
    requires_xf86vidmodeproto = ["xorg/xf86vidmodeproto"]
    requires_xorg_macros = ["xorg/macros"]
    requires_xproto = ["xorg/xproto"]


@attributes.requires("requires_git")
@attributes.requires("requires_xau")
@attributes.requires("requires_xorg_macros")
@attributes.requires("requires_xcb_proto")
@pkgconfig.to_cxxinfo(["xcb"])
@autotools.requires()
@libtool.relocate()
class Libxcb(autotools.Autotools):
    name = "xorg/libxcb"
    version = Parameter("1.17.0", help="Libxcb version.")

    requires_git = ["fetch:alias=src,url=https://www.x.org/releases/individual/lib/libxcb-{version}.tar.gz"]
    requires_xau = ["xorg/libxau"]
    requires_xorg_macros = ["xorg/macros"]
    requires_xcb_proto = ["xorg/xcb-proto"]
    srcdir = "{fetch[src]}/libxcb-{version}"


@attributes.requires("requires_inputproto")
@attributes.requires("requires_kbproto")
@attributes.requires("requires_libxau")
@attributes.requires("requires_libxcb")
@attributes.requires("requires_libxdmcp")
@attributes.requires("requires_src")
@attributes.requires("requires_xextproto")
@attributes.requires("requires_xorg_macros")
@attributes.requires("requires_xproto")
@attributes.requires("requires_xtrans")
@autotools.requires()
@pkgconfig.requires()
@pkgconfig.to_cxxinfo(["x11", "x11-xcb"])
@libtool.relocate()
class Libx11(autotools.Autotools):
    name = "xorg/libx11"
    version = Parameter("1.8.12", help="Libx11 version.")

    requires_inputproto = ["xorg/inputproto"]
    requires_kbproto = ["xorg/kbproto"]
    requires_libxau = ["xorg/libxau"]
    requires_libxcb = ["xorg/libxcb"]
    requires_libxdmcp = ["xorg/libxdmcp"]
    requires_src = ["fetch:alias=src,url=https://www.x.org/releases/individual/lib/libX11-{version}.tar.gz"]
    requires_xextproto = ["xorg/xextproto"]
    requires_xorg_macros = ["xorg/macros"]
    requires_xproto = ["xorg/xproto"]
    requires_xtrans = ["xorg/xtrans"]
    srcdir = "{fetch[src]}/libX11-{version}"


@attributes.requires("requires_libx11")
@attributes.requires("requires_libxcb")
@attributes.common_metadata()
@libtool.relocate()
class LibX11Xcb(Task):
    """
    Republishes libx11 and libxcb as libx11-xcb for compatibility.

    Mainly, prefix must be shared in order to build mesa.
    """
    name = "xorg/libx11-xcb"
    requires_libx11 = ["xorg/libx11"]
    requires_libxcb = ["xorg/libxcb"]
    selfsustained = True

    def run(self, deps, tools):
        self.x11 = deps["xorg/libx11"]
        self.xcb = deps["xorg/libxcb"]

    def publish(self, artifact, tools):
        with tools.cwd(self.x11.path):
            artifact.collect("*", symlinks=True)
        with tools.cwd(self.xcb.path):
            artifact.collect("*", symlinks=True)


TaskRegistry.get().add_task_class(Glproto)
TaskRegistry.get().add_task_class(Inputproto)
TaskRegistry.get().add_task_class(Kbproto)
TaskRegistry.get().add_task_class(Libx11)
TaskRegistry.get().add_task_class(LibX11Xcb)
TaskRegistry.get().add_task_class(Libxcb)
TaskRegistry.get().add_task_class(LibXdmcp)
TaskRegistry.get().add_task_class(Libxxf86vm)
TaskRegistry.get().add_task_class(Randrproto)
TaskRegistry.get().add_task_class(Renderproto)
TaskRegistry.get().add_task_class(Xau)
TaskRegistry.get().add_task_class(XcbKeysyms)
TaskRegistry.get().add_task_class(XcbProto)
TaskRegistry.get().add_task_class(Xext)
TaskRegistry.get().add_task_class(Xextproto)
TaskRegistry.get().add_task_class(Xf86vidmodeproto)
TaskRegistry.get().add_task_class(XorgMacros)
TaskRegistry.get().add_task_class(Xproto)
TaskRegistry.get().add_task_class(Xrandr)
TaskRegistry.get().add_task_class(Xrender)
TaskRegistry.get().add_task_class(Xtrans)
TaskRegistry.get().add_task_class(Libxshmfence)



from jolt import attributes, Alias, BooleanParameter, Download, Parameter
from jolt.pkgs import cmake, re2c
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.system
@attributes.common_metadata()
class NinjaBin(Download):
    name = "ninja/bin"
    version = Parameter("1.13.2", help="Ninja version.")
    url = ["https://github.com/ninja-build/ninja/releases/download/v{version}/ninja-{ninja_system}.zip"]
    collect = [{"files": "*", "dest": "bin/"}]

    @property
    def ninja_system(self):
        if self.system == "darwin":
            return "mac"
        if self.system == "windows":
            return "win"
        return self.system


@attributes.requires("requires_git")
@attributes.requires("requires_re2c")
@cmake.requires()
class NinjaSrc(cmake.CMake):
    name = "ninja/src"
    version = Parameter("1.13.2", help="Ninja version.")
    tests = BooleanParameter(False, help="Build tests.")
    requires_git = ["git:url=https://github.com/ninja-build/ninja.git,rev=v{version}"]
    requires_re2c = ["re2c"]
    srcdir = "{git[ninja]}"
    options = ["BUILD_TESTING={tests[ON,OFF]}"]


class Ninja(Alias):
    name = "ninja"
    version = Parameter("1.13.2", help="Ninja version.")
    requires = ["ninja/bin:version={version}"]


TaskRegistry.get().add_task_class(Ninja)
TaskRegistry.get().add_task_class(NinjaBin)
TaskRegistry.get().add_task_class(NinjaSrc)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class Zstd(cmake.CMake):
    name = "zstd"
    version = Parameter("ebc93b0", help="zstd version.")
    pic = BooleanParameter(False, "Build with position independent code.")
    shared = BooleanParameter(False, "Enable shared libraries.")
    requires_git = ["git:url=https://github.com/facebook/zstd.git,rev={version}"]
    srcdir = "{git[zstd]}"
    options = [
        "CMAKE_POSITION_INDEPENDENT_CODE={pic[ON,OFF]}",
        "ZSTD_BUILD_SHARED={shared[ON,OFF]}",
        "ZSTD_BUILD_STATIC={shared[OFF,ON]}",
        "ZSTD_BUILD_TESTS=OFF",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.system == "windows":
            artifact.cxxinfo.libraries.append("zstd_static")
        else:
            artifact.cxxinfo.libraries.append("zstd")


TaskRegistry.get().add_task_class(Zstd)




class JoltOptions(object):
    """ Jolt options that control the behavior of builds. """

    debug = False
    """ Enable debug mode. Break into debugger on exceptions. """

    download = True
    """ Enable downloading of remote artifacts, both session and persistent. """

    download_session = True
    """ Enable downloading of remote session artifacts. """

    keep_going = False
    """ Keep going with the build after a task fails. """

    local = False
    """ Disable network access. """

    network = False
    """ Distribute tasks to workers. """

    upload = True
    """ Enable uploading of artifacts. """

    worker = False
    """ Running as a worker. """

    salt = None
    """ Salt for hashing (--salt). """

    jobs = 1
    """ Number of concurrent local tasks to run (1). """

    mute = False
    """ Mute task output, until a task fails. """

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)



from jolt import attributes, Parameter
from jolt.pkgs import cmake
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
class GFlags(cmake.CMake):
    name = "gflags"
    version = Parameter("2.3.0", help="gflags version.")
    options = ["BUILD_SHARED_LIBS=ON"]
    requires_git = ["git:url=https://github.com/gflags/gflags.git,rev=v{version}"]
    srcdir = "{git[gflags]}"


TaskRegistry.get().add_task_class(GFlags)



import os
from jolt import attributes, Alias, Task, Parameter
from jolt.plugins import autotools, git, meson
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@autotools.requires()
class PkgConfPosix(autotools.Autotools):
    """ Package that provides the 'pkgconf' binary, an alternative implementation of 'pkg-config'. """

    name = "pkgconf/posix"
    version = Parameter("2.5.1", help="pkg-config version.")
    requires_git = ["git:url=https://github.com/pkgconf/pkgconf.git,rev=pkgconf-{version}"]
    srcdir = "{git[pkgconf]}"
    options = ["--enable-static", "--disable-shared"]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.PKG_CONFIG = "pkgconf"

        # Install pkg-config symlink
        with tools.cwd(artifact.path, "bin"):
            tools.symlink("pkgconf", "pkg-config")


@attributes.requires("requires_git")
@meson.requires()
class PkgConfWin32(meson.Meson):
    """ Package that provides the 'pkgconf' binary, an alternative implementation of 'pkg-config'. """

    name = "pkgconf/win32"
    version = Parameter("2.5.1", help="pkg-config version.")
    requires_git = ["git:url=https://github.com/pkgconf/pkgconf.git,rev=pkgconf-{version}"]
    srcdir = "{git[pkgconf]}"
    options = [
        "default_library=static",
        "tests=disabled",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.PKG_CONFIG = "pkgconf"

        # Install pkg-config copy
        with tools.cwd(artifact.path, "bin"):
            tools.copy("pkgconf", "pkg-config")


@attributes.requires("requires_{system}")
@attributes.system
class PkgConf(Alias):
    """ Alias for pkgconf package. """

    name = "pkgconf"
    version = Parameter("2.5.1", help="pkg-config version.")
    requires_darwin = ["pkgconf/posix:version={version}"]
    requires_linux = requires_darwin
    requires_windows = ["pkgconf/win32:version={version}"]


TaskRegistry.get().add_task_class(PkgConf)
TaskRegistry.get().add_task_class(PkgConfPosix)
TaskRegistry.get().add_task_class(PkgConfWin32)



from jolt.cache import ArtifactAttributeSetProvider


class StringVariableSet(object):
    """
    A set of string variables for an artifact.

    Example:

    .. code-block:: python

        artifact.strings.foo = "bar"
        print(artifact.strings.foo)

    """

    def __init__(self, artifact):
        super(StringVariableSet, self).__setattr__("_attributes", {})
        super(StringVariableSet, self).__setattr__("_artifact", artifact)

    def _get_attributes(self):
        return self._attributes

    def __getattr__(self, name):
        attributes = self._get_attributes()
        return attributes.get(name, None)

    def __setattr__(self, name, value):
        if not isinstance(value, str):
            raise ValueError(f"Value assigned to artifact.strings.{name} must be a string, got {type(value)}")
        attributes = self._get_attributes()
        attributes[name] = self._artifact.tools.expand(value)
        return value

    def __dict__(self):
        return {key: str(value) for key, value in self.items()}

    def items(self):
        return self._get_attributes().items()


@ArtifactAttributeSetProvider.Register
class StringVariableSetProvider(ArtifactAttributeSetProvider):
    """
    A provider of string variable sets.
    """

    def create(self, artifact):
        setattr(artifact, "strings", StringVariableSet(artifact))

    def parse(self, artifact, content):
        if "strings" not in content:
            return

        for key, value in content["strings"].items():
            setattr(artifact.strings, key, value)

    def format(self, artifact, content):
        if "strings" not in content:
            content["strings"] = {}

        for key, value in artifact.strings.items():
            content["strings"][key] = str(value)



from jolt import attributes, BooleanParameter, Parameter
from jolt.plugins import cmake, cxxinfo, git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["lz4"])
class Lz4(cmake.CMake):
    name = "lz4"
    version = Parameter("1.10.0", help="LZ4 version.")
    shared = BooleanParameter(False, help="Build shared libraries")
    requires_git = ["git:url=https://github.com/lz4/lz4.git,rev=v{version}"]
    srcdir = "{git[lz4]}/build/cmake"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
    ]


TaskRegistry.get().add_task_class(Lz4)



from jolt import attributes, Parameter, BooleanParameter
from jolt.pkgs import cmake, dbus, ninja, libglvnd, protobuf
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_dbus_{system}")
@attributes.requires("requires_git")
@attributes.requires("requires_{system}_gl_{gl}")
@attributes.requires("requires_ninja")
@attributes.requires("requires_protobuf")
@attributes.system
@cmake.options("options_{system}_gl_{gl}")
@cmake.requires()
@cmake.use_ninja()
class Qt(cmake.CMake):
    name = "qt"
    version = Parameter("6.10.1", help="Qt version.")
    gl = BooleanParameter(True, help="Enable OpenGL support.")
    webengine = BooleanParameter(False, help="Enable WebEngine module.")

    generator = "Ninja"
    requires_dbus_linux = ["dbus"]
    requires_git = ["git:path={buildroot}/git-qt,url=https://github.com/qt/qt5.git,rev=v{version},submodules=true"]
    requires_ninja = ["ninja"]
    requires_linux_gl_true = ["libglvnd"]
    requires_protobuf = ["protobuf"]
    srcdir = "{git[qt5]}"

    options = [
        "BUILD_qtwebengine={webengine[ON,OFF]}",
    ]


TaskRegistry.get().add_task_class(Qt)



from jolt import attributes
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.error import raise_error_if


class PkgConfigHelper(object):

    TEMPLATE_PKGCONFIG = """
prefix={{ artifact.final_path }}

Name: {{ pkgname }}
Description: {{ pkgname }} package
Version: {{ artifact.identity }}

Cflags: {% for flag in cxxflags %}{{ flag }} {% endfor %}{% for inc in incpaths %}-I${prefix}/{{ inc }} {% endfor %}{% for macro in macros %}-D{{ macro }} {% endfor %}

Libs: {% for flag in ldflags %}{{ flag }} {% endfor %}{% for libpath in libpaths %}-L${prefix}/{{ libpath }} {% endfor %}{% for library in libraries %}-l{{ library }} {% endfor %}

"""

    def __init__(self, artifact, tools):
        self.artifact = artifact
        self.tools = tools

    def _mkpath(self, path):
        if fs.path.commonpath([path, self.artifact.path]) != self.artifact.path:
            return path
        return fs.path.relpath(path, self.artifact.path)

    def cflags(self, package):
        package = " ".join(utils.as_list(package))
        try:
            with self.tools.environ(**self.environ):
                output = self.tools.run("{} --maximum-traverse-depth=1 --cflags-only-other {}", self.pkgconfig, package, output=False)
                return output.strip().split()
        except Exception as e:
            log.debug("PkgConfig.cflags: {}", str(e))
            return []

    def incpaths(self, package):
        package = " ".join(utils.as_list(package))
        try:
            with self.tools.environ(**self.environ):
                output = self.tools.run("{} --maximum-traverse-depth=1 --cflags-only-I {}", self.pkgconfig, package, output=False)
                return [self._mkpath(inc[2:]) for inc in output.strip().split()]
        except Exception as e:
            log.debug("PkgConfig.incpaths: {}", str(e))
            return []

    def linkflags(self, package):
        package = " ".join(utils.as_list(package))
        try:
            with self.tools.environ(**self.environ):
                output = self.tools.run("{} --maximum-traverse-depth=1 --libs-only-other {}", self.pkgconfig, package, output=False)
                return output.strip().split()
        except Exception as e:
            log.debug("PkgConfig.linkflags: {}", str(e))
            return []

    def libpaths(self, package):
        package = " ".join(utils.as_list(package))
        try:
            with self.tools.environ(**self.environ):
                output = self.tools.run("{} --maximum-traverse-depth=1 --libs-only-L {}", self.pkgconfig, package, output=False)
                return [self._mkpath(lib[2:]) for lib in output.strip().split()]
        except Exception as e:
            log.debug("PkgConfig.libpaths: {}", str(e))
            return []

    def libraries(self, package):
        package = " ".join(utils.as_list(package))
        try:
            with self.tools.environ(**self.environ):
                output = self.tools.run("{} --maximum-traverse-depth=1 --libs-only-l {}", self.pkgconfig, package, output=False)
                return [lib[2:] for lib in output.strip().split()]
        except Exception as e:
            log.debug("PkgConfig.libraries: {}", str(e))
            return []

    def write_pc(self, package):
        with self.tools.tmpdir() as tmpdir, self.tools.cwd(tmpdir):
            content = self.tools.render(
                self.TEMPLATE_PKGCONFIG,
                artifact=self.artifact,
                pkgname=package,
                macros=list(self.artifact.cxxinfo.macros),
                incpaths=list(self.artifact.cxxinfo.incpaths),
                libpaths=list(self.artifact.cxxinfo.libpaths),
                libraries=list(self.artifact.cxxinfo.libraries),
            )
            print(content)
            self.tools.write_file(f"{package}.pc", content, expand=False)
            self.artifact.collect(f"{package}.pc", "lib/pkgconfig/")

    @property
    def pkgconfig(self):
        pkgconf = self.tools.which(self.tools.getenv("PKG_CONFIG", "pkg-config"))
        raise_error_if(not pkgconf, "No pkg-config tool found in PATH")
        return pkgconf

    @property
    def environ(self):
        path = self.artifact.environ.get("PKG_CONFIG_PATH")
        if path is None:
            self.tools._task.verbose("No PKG_CONFIG_PATH in artifact environment")

        # Path from the artifact environment
        path = str(path).split(fs.pathsep)
        path = fs.pathsep.join(fs.path.join(self.artifact.path, p) for p in path)

        # Append existing PKG_CONFIG_PATH from the tools environment
        if self.tools.getenv("PKG_CONFIG_PATH"):
            path = path + fs.pathsep + self.tools.getenv("PKG_CONFIG_PATH")

        return {"PKG_CONFIG_PATH": path}


def to_cxxinfo(
    pkg: list | str,
    cflags: bool = True,
    cxxflags: bool = True,
    incpaths: bool = True,
    ldflags: bool = True,
    libpaths: bool = True,
    libraries: bool = True,
):
    """
    Decorator to add pkg-config information to cxxinfo metadata of an artifact.

    The decorator enables interoperability between libraries built with Jolt's
    Ninja plugin and external packages that provide pkg-config files.

    It uses the pkg-config tool to query for compiler and linker flags,
    include paths, library paths, and libraries associated with a given package.
    If the relevant flags are found, they are appended to the corresponding fields
    in the artifact's cxxinfo metadata.

    Args:
        pkg (str): The name of the pkg-config package to be added to cxxinfo.
        cflags (bool): Whether to add C compiler flags from pkg-config.
        cxxflags (bool): Whether to add C++ compiler flags from pkg-config.
        incpaths (bool): Whether to add include paths from pkg-config.
        ldflags (bool): Whether to add linker flags from pkg-config.
        libpaths (bool): Whether to add library paths from pkg-config.
        libraries (bool): Whether to add libraries from pkg-config.
    """

    def decorate(cls):
        original_publish = cls.publish

        def publish(self, artifact, tools):
            original_publish(self, artifact, tools)

            pc = PkgConfigHelper(artifact, tools)
            if not pc.environ:
                self.verbose("Skipping pkg-config cxxinfo addition due to missing PKG_CONFIG_PATH.")
                return

            if cflags:
                artifact.cxxinfo.cflags.extend(pc.cflags(pkg))
            if cxxflags:
                artifact.cxxinfo.cxxflags.extend(pc.cflags(pkg))
            if incpaths:
                artifact.cxxinfo.incpaths.extend(pc.incpaths(pkg))
            if ldflags:
                artifact.cxxinfo.ldflags.extend(pc.linkflags(pkg))
            if libpaths:
                artifact.cxxinfo.libpaths.extend(pc.libpaths(pkg))
            if libraries:
                artifact.cxxinfo.libraries.extend(pc.libraries(pkg))

        cls.publish = publish
        return cls

    return decorate


def from_cxxinfo(package):
    """
    Decorator to write a pkg-config file for the given package
    based on the cxxinfo metadata of the artifact.
    """

    def decorate(cls):
        original_publish = cls.publish
        original_unpack = cls.unpack

        def publish(self, artifact, tools):
            original_publish(self, artifact, tools)

            pc = PkgConfigHelper(artifact, tools)
            pc.write_pc(package)

        def unpack(self, artifact, tools):
            original_unpack(self, artifact, tools)

            pc = PkgConfigHelper(artifact, tools)
            pc.write_pc(package)

        cls.publish = publish
        cls.unpack = unpack
        return cls

    return decorate


def requires():
    """ Decorator to add pkg-config requirements to a task. """

    import jolt.pkgs.pkgconfig

    def decorate(cls):
        cls = attributes.requires("requires_pkgconf")(cls)
        cls.requires_pkgconf = ["pkgconf"]
        return cls

    return decorate



from jolt import attributes, Parameter
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
class HDF5(cmake.CMake):
    name = "hdf5"
    version = Parameter("2.0.0", help="hdf5 version.")
    requires_git = ["git:url=https://github.com/HDFGroup/hdf5.git,rev={version}"]
    srcdir = "{git[hdf5]}"
    options = [
        "BUILD_SHARED_LIBS=ON",
        "HDF5_BUILD_EXAMPLES=OFF",
    ]


TaskRegistry.get().add_task_class(HDF5)



from jolt import attributes, BooleanParameter, Parameter, Task
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["Catch2"])
class Catch2(cmake.CMake):
    name = "catch2"
    version = Parameter("3.11.0", help="catch2 version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/catchorg/Catch2.git,rev=v{version}"]
    srcdir = "{git[Catch2]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
    ]


class Catch2Main(Task):
    name = "catch2/main"
    extends = "catch2"

    def publish(self, artifact, tools):
        artifact.cxxinfo.libraries.append("Catch2Main")


TaskRegistry.get().add_task_class(Catch2)
TaskRegistry.get().add_task_class(Catch2Main)



import os
from jolt import attributes, Parameter
from jolt.pkgs import cmake, openssl
from jolt.plugins import cxxinfo, git, cmake, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_ssl")
@cmake.requires()
@cmake.use_ninja()
@pkgconfig.requires()
@cxxinfo.publish(libraries=["event"])
class LibEvent(cmake.CMake):
    name = "libevent"
    version = Parameter("2.1.12", help="libevent version.")
    options = ["CMAKE_POLICY_VERSION_MINIMUM=3.5"]
    requires_git = ["git:url=https://github.com/libevent/libevent.git,rev=release-{version}-stable"]
    requires_ssl = ["openssl"]
    srcdir = "{git[libevent]}"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)

        for libdir in ["lib", "lib32", "lib64"]:
            if os.path.isdir(os.path.join(artifact.path, libdir)):
                self.libdir = libdir
                break

        with tools.cwd(artifact.path, "{libdir}/cmake/libevent"):
            for file in tools.glob("*.cmake"):
                tools.replace_in_file(
                    file,
                    artifact.strings.install_prefix,
                    "${{CMAKE_CURRENT_LIST_DIR}}/../../..",
                )


TaskRegistry.get().add_task_class(LibEvent)



from jolt import BooleanParameter, Parameter, attributes
from jolt.pkgs import cmake
from jolt.plugins import cmake, cxxinfo, git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class SDL(cmake.CMake):
    name = "sdl"
    version = Parameter("3.2.28", help="SDL version.")
    shared = BooleanParameter(False, help="Build shared library")
    requires_git = ["git:url=https://github.com/libsdl-org/SDL.git"]
    srcdir = "{git[SDL]}"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.system == "windows":
            artifact.cxxinfo.libraries.append("SDL2{shared[,-static]}")
        else:
            artifact.cxxinfo.libraries.append("SDL2")


TaskRegistry.get().add_task_class(SDL)



from jolt import attributes, BooleanParameter, Parameter
from jolt.plugins import cmake, git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
class Blake3(cmake.CMake):
    name = "blake3"
    version = Parameter("1.8.3", help="blake3 version.")
    pic = BooleanParameter(False, help="Build with position independent code.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/BLAKE3-team/BLAKE3.git,rev={version}"]
    srcdir = "{git[BLAKE3]}/c"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "CMAKE_POSITION_INDEPENDENT_CODE={pic[ON,OFF]}",
    ]


TaskRegistry.get().add_task_class(Blake3)



from jolt import attributes, BooleanParameter, Parameter, Task
from jolt.pkgs import abseil, cares, cmake, nasm, protobuf, re2, ssl, zlib
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_abseil")
@attributes.requires("requires_cares")
@attributes.requires("requires_git")
@attributes.requires("requires_nasm_{system}")
@attributes.requires("requires_protobuf")
@attributes.requires("requires_re2")
@attributes.requires("requires_ssl")
@attributes.requires("requires_zlib")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class Grpc(cmake.CMake):
    name = "grpc"
    version = Parameter("1.76.0", help="Grpc version.")
    pic = BooleanParameter(False, help="Build with position independent code.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    generator = "Ninja"
    requires_abseil = ["abseil:pic={pic},shared={shared}"]
    requires_cares = ["c-ares:pic={pic},shared={shared}"]
    requires_git = ["git:url=https://github.com/grpc/grpc.git,path={buildroot}/git_grpc,rev=v{version},submodules=true"]
    requires_nasm_windows = ["nasm"]
    requires_protobuf = ["protobuf:pic={pic},shared={shared}"]
    requires_re2 = ["re2:pic={pic},shared={shared}"]
    requires_ssl = ["virtual/ssl:pic={pic},shared={shared}"]
    requires_zlib = ["virtual/zlib:pic={pic},shared={shared}"]
    srcdir = "{git[grpc]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "CMAKE_POSITION_INDEPENDENT_CODE={pic[ON,OFF]}",
        "CMAKE_CXX_STANDARD=17",
        "gRPC_ABSL_PROVIDER=package",
        "gRPC_BUILD_TESTS=OFF",
        "gRPC_CARES_PROVIDER=package",
        "gRPC_MSVC_STATIC_RUNTIME=OFF",
        "gRPC_PROTOBUF_PROVIDER=package",
        "gRPC_RE2_PROVIDER=package",
        "gRPC_SSL_PROVIDER=package",
        "gRPC_ZLIB_PROVIDER=package",
        "gRPC_USE_SYSTEMD=OFF",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.system == "windows":
            artifact.cxxinfo.msvcrt = "Dynamic"


class GrpcC(Task):
    name = "grpc/c"
    extends = "grpc"

    def publish(self, artifact, tools):
        artifact.cxxinfo.libraries.append("grpc")


class GrpcCXX(Task):
    name = "grpc/c++"
    extends = "grpc"

    def publish(self, artifact, tools):
        artifact.cxxinfo.libraries.append("grpc++")


TaskRegistry.get().add_task_class(Grpc)
TaskRegistry.get().add_task_class(GrpcC)
TaskRegistry.get().add_task_class(GrpcCXX)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import libcap
from jolt.plugins import git, autotools, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_libcap_{caps[on,off]}")
@autotools.requires()
@pkgconfig.requires()
class CRun(autotools.Autotools):
    name = "crun"
    version = Parameter("1.26", help="crun version.")
    caps = BooleanParameter(False, help="Enable capabilities support.")
    systemd = BooleanParameter(False, help="Enable systemd support.")
    requires_git = ["git:url=https://github.com/containers/crun.git,rev={version},submodules=true"]
    requires_caps_on = ["libcap"]
    srcdir = "{git[crun]}"
    options = [
        # Unable to get ./configure to pick these up from pkg-config
        "--disable-criu",
        "--disable-seccomp",
        "--enable-embedded-yajl",
        "{caps[,--disable-caps]}",
        "{systemd[,--disable-systemd]}",
    ]


TaskRegistry.get().add_task_class(CRun)



from jolt import attributes, Parameter
from jolt.pkgs import gperf, libtool
from jolt.plugins import git, autotools, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_gperf")
@attributes.requires("requires_libtool")
@pkgconfig.requires()
@pkgconfig.to_cxxinfo("libseccomp")
class Libseccomp(autotools.Autotools):
    name = "libseccomp"
    version = Parameter("2.5.6", help="libseccomp version.")
    requires_git = ["git:url=https://github.com/seccomp/libseccomp.git,rev=v{version}"]
    requires_gperf = ["gperf"]
    requires_libtool = ["libtool"]
    srcdir = "{git[libseccomp]}"


TaskRegistry.get().add_task_class(Libseccomp)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
class SpirvTools(cmake.CMake):
    name = "spirv-tools"
    version = Parameter("2024.4", help="SPIRV-Tools version.")
    tests = BooleanParameter(False, help="Build tests.")
    requires_git = ["git:url=https://github.com/KhronosGroup/SPIRV-Tools.git,rev=v{version}"]
    srcdir = "{git[SPIRV-Tools]}"
    options = ["SPIRV_SKIP_TESTS={tests[OFF,ON]}"]

    def run(self, deps, tools):
        with tools.cwd(self.srcdir):
            tools.run("python3 utils/git-sync-deps")
        super().run(deps, tools)


TaskRegistry.get().add_task_class(SpirvTools)



from jolt import attributes, Parameter
from jolt.plugins import autotools, fetch
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_texinfo")
@autotools.requires(automake=False, libtool=False)
class GnuMake(autotools.Autotools):
    name = "make"
    version = Parameter("4.4.1", help="Automake version.")
    requires_git = ["fetch:alias=src,url=https://ftpmirror.gnu.org/gnu/make/make-{version}.tar.gz"]
    requires_texinfo = ["texinfo"]
    srcdir = "{fetch[src]}/make-{version}"


TaskRegistry.get().add_task_class(GnuMake)



from jolt import attributes, Parameter
from jolt.pkgs import help2man, texinfo
from jolt.plugins import autotools, git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_help2man")
@attributes.requires("requires_texinfo")
@autotools.requires(libtool=False)
class Libtool(autotools.Autotools):
    name = "libtool"
    version = Parameter("2.6.0", help="Libtool version.")

    requires_git = ["git:url=git://git.git.savannah.gnu.org/libtool.git,rev=v{version}"]
    requires_help2man = ["help2man"]
    requires_texinfo = ["texinfo"]
    srcdir = "{git[libtool]}"

    def run(self, deps, tools):
        with tools.cwd("{git[libtool]}"):
            tools.run("./bootstrap")
        super().run(deps, tools)

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.LIBTOOL = "libtool"
        artifact.environ.ACLOCAL_PATH.append("share/libtool")

    def unpack(self, artifact, tools):
        with tools.cwd(artifact.path):
            tools.replace_in_file(
                "bin/libtoolize",
                artifact.strings.install_prefix,
                artifact.path)
            tools.replace_in_file(
                "lib/libltdl.la",
                artifact.strings.install_prefix,
                artifact.path)


TaskRegistry.get().add_task_class(Libtool)



from jolt import attributes
from jolt import Download
from jolt import Parameter
from jolt.tasks import TaskRegistry

import platform


@attributes.system
class NodeJS(Download):
    name = "nodejs"
    version = Parameter("25.2.1", help="NodeJS version.")
    url = "https://nodejs.org/dist/v{version}/node-v{version}-{os}-{arch}.{ext}"
    bin = "node-v{version}-{os}-{arch}/bin"

    @property
    def os(self):
        if self.system == "windows":
            return "win"
        return self.system

    @property
    def ext(self):
        if self.system == "windows":
            return "zip"
        return "tar.gz"

    @property
    def arch(self):
        arch = platform.machine().lower()
        if arch in ["amd64", "x86_64"]:
            return "x64"
        if arch in ["i386", "i486", "i586", "i686"]:
            return "x86"
        if arch in ["arm", "armv7l"]:
            return "armv7l"
        if arch in ["arm64", "armv8l", "aarch64"]:
            return "arm64"
        return None

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.PATH.append(self.bin)


TaskRegistry.get().add_task_class(NodeJS)



from jolt import attributes, BooleanParameter, Parameter, Task
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_abseil")
@attributes.requires("requires_git")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class Protobuf(cmake.CMake):
    """
    Protocol buffers main package.

    For linking, require protobuf/lib and friends.
    """

    name = "protobuf"
    version = Parameter("33.2", help="Protobuf version.")
    pic = BooleanParameter(False, help="Build with position independent code.")
    shared = BooleanParameter(False, help="Build shared libraries")
    requires_abseil = ["abseil:pic={pic},shared={shared}"]
    requires_git = ["git:url=https://github.com/protocolbuffers/protobuf.git,rev=v{version}"]
    srcdir = "{git[protobuf]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "CMAKE_CXX_STANDARD=17",
        "CMAKE_POSITION_INDEPENDENT_CODE={pic[ON,OFF]}",
        "protobuf_BUILD_TESTS=OFF",
        "protobuf_MSVC_STATIC_RUNTIME=OFF",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.system == "windows":
            artifact.cxxinfo.msvcrt = "Dynamic"


class ProtobufLib(Task):
    """ Protobuf library """

    name = "protobuf/lib"
    version = Parameter("33.2", help="Protobuf version.")
    shared = BooleanParameter(False, help="Build shared libraries")
    extends = "protobuf:version={version},shared={shared}"

    def publish(self, artifact, tools):
        artifact.cxxinfo.libraries.append("protobuf")


class ProtobufLite(Task):
    """ Protobuf Lite library """

    name = "protobuf/lib-lite"
    version = Parameter("33.2", help="Protobuf version.")
    shared = BooleanParameter(False, help="Build shared libraries")
    extends = "protobuf:version={version},shared={shared}"

    def publish(self, artifact, tools):
        artifact.cxxinfo.libraries.append("protobuf-lite")


class ProtobufLibProtoc(Task):
    """ Protobuf compiler library (libprotoc) """

    name = "protobuf/lib-compiler"
    version = Parameter("33.2", help="Protobuf version.")
    shared = BooleanParameter(False, help="Build shared libraries")
    extends = "protobuf:version={version},shared={shared}"

    def publish(self, artifact, tools):
        artifact.cxxinfo.libraries.append("protoc")


TaskRegistry.get().add_task_class(Protobuf)
TaskRegistry.get().add_task_class(ProtobufLib)
TaskRegistry.get().add_task_class(ProtobufLite)
TaskRegistry.get().add_task_class(ProtobufLibProtoc)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class Libxml2(cmake.CMake):
    name = "libxml2"
    version = Parameter("2.15.1", help="Libxml2 version.")
    shared = BooleanParameter(False, "Enable shared libraries.")
    requires_git = ["git:url=https://gitlab.gnome.org/GNOME/libxml2.git,rev=v{version}"]

    srcdir = "{git[libxml2]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "LIBXML2_WITH_ICONV=OFF",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.system == "windows":
            artifact.cxxinfo.libraries.append("libxml2{shared[,s]}")
        else:
            artifact.cxxinfo.libraries.append("xml2")
            artifact.cxxinfo.libraries.append("m")
            artifact.cxxinfo.libraries.append("dl")


TaskRegistry.get().add_task_class(Libxml2)



from jolt import attributes, Parameter, Task
from jolt.tasks import TaskRegistry


@attributes.system
class Fstree(Task):
    name = "fstree"
    version = Parameter("24.11.128")

    def publish(self, artifact, tools):
        self.ext = ".exe" if self.system == "windows" else ""
        with tools.cwd(tools.builddir()):
            tools.mkdir("jolt/bin")
            tools.download("https://github.com/srand/fstree/releases/download/{version}/fstree-{system}-x86_64{ext}", "jolt/bin/fstree-{system}-x86_64{ext}")
            tools.chmod("jolt/bin/fstree-{system}-x86_64{ext}", 0o755)
            artifact.collect("jolt/bin/fstree-{system}-x86_64{ext}", "bin/fstree{ext}")
            artifact.environ.PATH.append("bin")


TaskRegistry.get().add_task_class(Fstree)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class JsonCPP(cmake.CMake):
    name = "jsoncpp"
    version = Parameter("1.9.6", help="JsonCPP version.")
    tests = BooleanParameter(False, help="Build tests.")
    requires_git = ["git:url=https://github.com/open-source-parsers/jsoncpp.git,rev={version}"]
    srcdir = "{git[jsoncpp]}"
    options = [
        "JSONCPP_WITH_TESTS={tests[ON,OFF]}",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.system == "windows":
            artifact.cxxinfo.libraries.append("jsoncpp_static")
        else:
            artifact.cxxinfo.libraries.append("jsoncpp")


TaskRegistry.get().add_task_class(JsonCPP)
